# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _
from ..parsed_tf_node import ParsedTFNode
from ..basic_graph_ops import *  # pylint: disable=unused-wildcard-import,wildcard-import
from ..tfssa import SSAFunction
from .visitors import (
    FindAllReachableNodes,
    FindImmediateDownstreamNodes,
    FindImmediateUpstreamNodes,
    FindSubgraph,
)
import logging


class FunctionalizeLoops(object):
    """
    Turns while loops in TensorFlow dataflow graph into the functional form:
    while(cond_function, body_function)

    Usage:
    Given a graph in tfssa (the NetworkEnsemble defined in network.py) form:

    This will functionalize *ONE* loop in the main function.

        f = FunctionalizeLoops()
        ret = f.functionalize_loops(self, tfssa, "main")

    if ret is True, one loop has been functionalized, and the new functions
    added to tfssa. If False, there is no loop to functionalize.

    Generally, repeated calls to this will be necessary to catch all loops.

    Instead, use functionalize_loops.
    """

    def __init__(self):
        self.exits = None
        self.merges = None
        self.enters = None
        self.constant_enters = None
        self.switches = None
        self.subgraph = None
        self.loopcond = None
        self.is_constant = None
        self.next_iterations = None
        self.cond = None
        self.body = None

    def _search(self, g, node):
        if not isinstance(node, ParsedTFNode):
            node = g[node]

        # we look for NextIteration nodes
        assert node.op == "Enter"

        frame_name = node.attr["frame_name"]
        logging.debug("Fixing frame name: %s", frame_name)
        # find all the enter args
        # this is basically the enter frame
        # functionalize_control_flow.cc:FunctionalizeControlFlow (1160-1196)
        self.enters = [
            k for k, v in g.items() if v.attr.get("frame_name", "") == frame_name
        ]
        self.is_constant = [
            bool(g[n].attr.get("is_constant", False)) for n in self.enters
        ]
        self.merges = (
            FindImmediateDownstreamNodes(lambda x: x.op == "Merge")
            .visit_many(g, self.enters)
            .get_result()
        )
        self.next_iterations = (
            FindImmediateUpstreamNodes(lambda x: x.op == "NextIteration")
            .visit_many(g, self.merges)
            .get_result()
        )
        self.switches = (
            FindImmediateDownstreamNodes(lambda x: x.op == "Switch")
            .visit_many(g, self.merges)
            .get_result()
        )
        self.exits = (
            FindImmediateDownstreamNodes(lambda x: x.op == "Exit")
            .visit_many(g, self.switches)
            .get_result()
        )
        self.loopcond = list(
            set(
                FindImmediateUpstreamNodes(lambda x: x.op == "LoopCond")
                .visit_many(g, self.switches)
                .get_result()
            )
        )

        self.subgraph = FindSubgraph(self.exits).visit_many(g, self.enters).get_result()
        self.cond = FindSubgraph(self.switches).visit_many(g, self.merges).get_result()
        self.body = (
            FindSubgraph([node.name] + self.exits)
            .visit_many(g, self.switches)
            .get_result()
        )
        # drop merges and switches from cond and body
        self.cond = [
            i for i in self.cond if i not in (self.merges + self.switches + self.enters)
        ]
        self.body = (
            [i for i in self.body if i not in ([node.name] + self.switches)]
            + [node.name]
            + self.switches
            + self.merges
            + self.enters
        )

        # ok. we can now rebuild.

    def _fix_graph_invariants(self, g):
        import copy

        check = lambda x: x is not None and len(x) > 0
        check(self.exits)
        check(self.merges)
        check(self.enters)
        check(self.switches)
        check(self.subgraph)
        check(self.cond)
        check(self.loopcond)
        assert len(self.loopcond) == 1
        # maintain the invariant of a unique Enter node per argument
        # functionalize_control_flow.cc:FunctionalizeLoop (295)
        for i in copy.copy(self.enters):
            node = g[i]
            assert len(node.outputs) > 0
            assert len(node.inputs) == 1
            assert len(node.control_inputs) == 0
            assert len(node.control_outputs) == 0
            if len(node.outputs) == 1:
                continue
            node_output_copy = copy.copy(node.outputs)
            for j in range(1, len(node_output_copy)):
                # make a new enter node for each
                new_enter_node = copy.deepcopy(node)
                new_enter_node.inputs = []
                new_enter_node.outputs = []
                new_enter_node.name = node.name + "/trsplit%d" % (j)
                g[new_enter_node.name] = new_enter_node
                logging.debug("splitting %s", node.name)
                # connect the new node
                enter_output = node_output_copy[j]
                disconnect_edge(g, node.name, enter_output)
                connect_edge(g, new_enter_node.name, enter_output)
                connect_sources(g, node.inputs, new_enter_node.name)
                # insert into graph
                self.enters.append(new_enter_node.name)

    def functionalize_loops(self, tfssa, function_to_functionalize):
        g = tfssa.functions[function_to_functionalize].graph
        loopni = [a for a in g if g[a].op == "Enter"]
        if len(loopni) == 0:
            return False
        self._search(g, loopni[0])

        self.constant_enters = [
            self.enters[i] for i in range(len(self.enters)) if self.is_constant[i]
        ]
        self.enters = [
            self.enters[i] for i in range(len(self.enters)) if not self.is_constant[i]
        ]
        self._fix_graph_invariants(g)
        # for each enter node, find the corresponding downstream merge node
        enter_corresponding_merge = [
            FindImmediateDownstreamNodes(lambda x: x.op == "Merge")
            .visit(g, enter)
            .get_result()[0]
            for enter in self.enters
        ]
        merge_corresponding_ni = [
            FindImmediateUpstreamNodes(lambda x: x.op == "NextIteration")
            .visit(g, merge)
            .get_result()[0]
            for merge in enter_corresponding_merge
        ]
        switch_corresponding_merge = []
        for merge in enter_corresponding_merge:
            switch_after_merge = (
                FindImmediateDownstreamNodes(lambda x: x.op == "Switch")
                .visit(g, merge)
                .get_result()
            )
            if len(switch_after_merge) > 0:
                switch_corresponding_merge.append(switch_after_merge[0])
            else:
                # There are some situations there is no switch not for a given
                # merge. While odd... its ok. we construct one
                # In this situation there is no Exit either, but it can be
                # constructed later on
                new_switch_node = ParsedTFNode()
                new_switch_node.op = "Switch"
                new_switch_node.name = tfssa._find_free_name("fake_switch_")
                g[new_switch_node.name] = new_switch_node
                connect_edge(g, merge, new_switch_node.name)
                connect_edge(g, self.loopcond[0], new_switch_node.name)
                switch_corresponding_merge.append(new_switch_node.name)

        exit_corresponding_switch = []
        for switch in switch_corresponding_merge:
            res = (
                FindImmediateDownstreamNodes(lambda x: x.op == "Exit")
                .visit(g, switch)
                .get_result()
            )
            if len(res) > 0:
                exit_corresponding_switch.append(res[0])
            else:
                new_exit_node = ParsedTFNode()
                new_exit_node.op = "Exit"
                new_exit_node.name = tfssa._find_free_name("fake_exit_")
                g[new_exit_node.name] = new_exit_node
                connect_edge(g, switch, new_exit_node.name)
                exit_corresponding_switch.append(new_exit_node.name)

        while_loop = ParsedTFNode()
        while_loop.op = "while"
        while_loop.name = tfssa._find_free_name("while_")
        g[while_loop.name] = while_loop

        # Build the Loop Condition

        # replace all enters with a single make_tuple
        # we replace merge with get_tuple and turn it into a function call
        # terminated with LoopCond
        make_inputs = ParsedTFNode()
        make_inputs.op = "make_tuple"
        make_inputs.name = tfssa._find_free_name("make_input_")
        g[make_inputs.name] = make_inputs
        for enter in self.enters:
            replace_dest(g, g[enter].inputs[0], enter, make_inputs.name)
        constant_base_index = len(make_inputs.inputs)
        for enter in self.constant_enters:
            replace_dest(g, g[enter].inputs[0], enter, make_inputs.name)

        connect_edge(g, make_inputs.name, while_loop.name)
        connect_dests(g, while_loop.name, exit_corresponding_switch)

        # build the cond function
        cond_body = ParsedTFNode()
        cond_body.op = "function_entry"
        cond_body.name = tfssa._find_free_name("cond_function_")
        cond_body.inputs = []
        g[cond_body.name] = cond_body
        for merge_idx in range(len(enter_corresponding_merge)):
            merge = enter_corresponding_merge[merge_idx]
            switch = switch_corresponding_merge[merge_idx]
            enter_node = g[self.enters[merge_idx]]
            merge_node = g[merge]
            if switch is not None:
                switch_node = g[switch]
            else:
                switch_node = None
            merge_node.op = "get_tuple"
            merge_node.attr = {"index": merge_idx}
            # disconnect merge from switch
            # disconnect loopcond from switch
            disconnect_edge(g, enter_node.name, merge_node.name)
            if switch_node is not None:
                disconnect_edge(g, merge_node.name, switch_node.name)
                disconnect_edge(g, self.loopcond[0], switch_node.name)
            for i in merge_node.inputs[:]:
                disconnect_edge(g, i, merge_node.name)
            connect_edge(g, cond_body.name, merge_node.name)
            # delete get_tuple if it does nothing
            if len(merge_node.outputs) == 0:
                delete_node(g, merge)

        g[self.loopcond[0]].op = "return"

        # build the body function
        body = ParsedTFNode()
        body.op = "function_entry"
        body.name = tfssa._find_free_name("body_function_")
        body.inputs = []
        g[body.name] = body
        for switch_idx in range(len(switch_corresponding_merge)):
            switch = switch_corresponding_merge[switch_idx]
            exit = exit_corresponding_switch[switch_idx]
            disconnect_edge(g, switch, exit)

            # replace switch with a get_tuple
            switch_node = g[switch]
            switch_node.op = "get_tuple"
            switch_node.attr = {"index": switch_idx}
            connect_edge(g, body.name, switch_node.name)
            # delete get_tuple if it does nothing
            if len(switch_node.outputs) == 0:
                delete_node(g, switch)

        # replace all next_iteration with a single make_tuple
        # we replace merge with get_tuple and turn it into a function call
        # terminated with LoopCond
        make_outputs = ParsedTFNode()
        make_outputs.op = "make_tuple"
        make_outputs.name = tfssa._find_free_name("make_output_")
        g[make_outputs.name] = make_outputs
        for ni in merge_corresponding_ni:
            connect_edge(g, g[ni].inputs[0], make_outputs.name)

        # connect constant enters to come from function
        # connect constant enters to exit
        for idx, enter in enumerate(self.constant_enters):
            body_connected = False
            for output in list(g[enter].outputs):
                if output not in self.cond and output not in self.body:
                    cond_intersection = (
                        FindSubgraph(self.cond).visit(g, output).get_result()
                    )
                    body_intersection = (
                        FindSubgraph(self.body).visit(g, output).get_result()
                    )
                    if len(cond_intersection) > 0:
                        cond_intersection.append(output)
                        self.cond += cond_intersection
                    if len(body_intersection) > 0:
                        body_intersection.append(output)
                        self.body += body_intersection
                get_tuple = ParsedTFNode()
                get_tuple.op = "get_tuple"
                get_tuple.name = tfssa._find_free_name("get_tuple_const_")
                get_tuple.attr = {"index": idx + constant_base_index}
                g[get_tuple.name] = get_tuple

                if output in self.cond:
                    connect_edge(g, cond_body.name, get_tuple.name)
                elif output in self.body:
                    connect_edge(g, body.name, get_tuple.name)
                replace_source(g, enter, output, get_tuple.name)

            # body must accept and return everything
            get_tuple = ParsedTFNode()
            get_tuple.op = "get_tuple"
            get_tuple.name = tfssa._find_free_name("get_tuple_const_")
            get_tuple.attr = {"index": idx + constant_base_index}
            g[get_tuple.name] = get_tuple
            connect_edge(g, body.name, get_tuple.name)
            connect_edge(g, get_tuple.name, make_outputs.name)

        assert len(g[make_outputs.name].inputs) == len(g[make_inputs.name].inputs)

        output_return = ParsedTFNode()
        output_return.op = "return"
        output_return.name = tfssa._find_free_name("body_return_")
        g[output_return.name] = output_return
        connect_edge(g, make_outputs.name, output_return.name)
        while_loop.attr["cond_function"] = cond_body.name
        while_loop.attr["body_function"] = body.name
        for i in self.enters:
            delete_node(g, i)
        for i in self.next_iterations:
            delete_node(g, i)
        for i in self.constant_enters:
            delete_node(g, i)

        for i in range(len(exit_corresponding_switch)):
            exit_node = exit_corresponding_switch[i]
            g[exit_node].op = "get_tuple"
            g[exit_node].attr = {"index": i}
        cond_function = (
            FindSubgraph(self.loopcond[0]).visit(g, cond_body.name).get_result()
        )
        cond_function = set(cond_function + [self.loopcond[0], cond_body.name])
        body_function = (
            FindSubgraph(output_return.name).visit(g, body.name).get_result()
        )
        body_function = set(body_function + [body.name, output_return.name])

        # trace input constants associated with the cond_graph
        # and the body_graph. These constants can only have one consumer
        # for now. Any more and we will either need to associate
        # it as an argument, or split the constant.
        cond_constants = (
            FindImmediateUpstreamNodes(lambda x: x.op == "Const")
            .visit_many(g, cond_function)
            .get_result()
        )
        body_constants = (
            FindImmediateUpstreamNodes(lambda x: x.op == "Const")
            .visit_many(g, body_function)
            .get_result()
        )
        # for const_node in cond_constants + body_constants:
        #    assert(len(g[const_node].outputs) == 1)

        cond_function = cond_function.union(set(cond_constants))
        body_function = body_function.union(set(body_constants))

        downstream_cond = (
            FindAllReachableNodes(lambda x: True)
            .visit_many(g, cond_function)
            .get_result()
        )
        downstream_cond = set(downstream_cond) - cond_function
        if len(downstream_cond) > 0:
            logging.debug(
                "Disconnecting unused variables in condition function %s",
                downstream_cond,
            )
            for i in downstream_cond:
                delete_node(g, i)

        downstream_body = (
            FindAllReachableNodes(lambda x: True)
            .visit_many(g, body_function)
            .get_result()
        )
        downstream_body = set(downstream_body) - body_function
        if len(downstream_body) > 0:
            logging.debug(
                "Disconnecting unused variables in body function %s", downstream_body
            )
            for i in downstream_body:
                delete_node(g, i)

        cond_graph = {k: v for k, v in g.items() if k in cond_function}
        body_graph = {k: v for k, v in g.items() if k in body_function}
        g = {
            k: v
            for k, v in g.items()
            if k not in cond_function and k not in body_function
        }
        # localize control dependencies
        # In the main graph, reattach the control dependency to the while op
        for k, v in g.items():
            for idx in range(len(v.control_inputs)):
                if v.control_inputs[idx] not in g:
                    v.control_inputs[idx] = while_loop.name
                    while_loop.control_outputs.append(k)
            for idx in range(len(v.control_outputs)):
                if v.control_outputs[idx] not in g:
                    v.control_outputs[idx] = while_loop.name
                    while_loop.control_inputs.append(k)

        # in the cond and body graphs, drop non-local control dependencies
        # entirely
        for graph in [cond_graph, body_graph]:
            for k, v in graph.items():
                for idx in range(len(v.control_inputs) - 1, -1, -1):
                    if v.control_inputs[idx] not in graph:
                        v.control_inputs.pop(idx)

                for idx in range(len(v.control_outputs) - 1, -1, -1):
                    if v.control_outputs[idx] not in graph:
                        v.control_outputs.pop(idx)
        tfssa.functions[function_to_functionalize] = SSAFunction(g)
        tfssa.add_function(cond_body.name, SSAFunction(cond_graph))
        tfssa.add_function(body.name, SSAFunction(body_graph))
        return True


def functionalize_loops(tfssa):
    """
    Functionalize all loops in an tfssa
    """
    done = False
    while not done:
        done = True
        for f in list(tfssa.functions.keys()):
            functionalize = FunctionalizeLoops()
            ret = functionalize.functionalize_loops(tfssa, f)
            if ret:
                done = False



# Copyright (c) 2017-2019, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause
from . import _utils
import logging
import keras as _keras
import numpy as _np
from ...proto import NeuralNetwork_pb2 as _NeuralNetwork_pb2

from distutils.version import StrictVersion as _StrictVersion

if _keras.__version__ >= _StrictVersion("2.2.1"):
    from keras.layers import DepthwiseConv2D
elif _keras.__version__ >= _StrictVersion("2.2.0"):
    from keras.layers import DepthwiseConv2D
    from keras_applications.mobilenet import relu6
else:
    from keras.applications.mobilenet import DepthwiseConv2D, relu6


def _get_recurrent_activation_name_from_keras(activation):
    if activation == _keras.activations.sigmoid:
        activation_str = "SIGMOID"
    elif activation == _keras.activations.hard_sigmoid:
        activation_str = "SIGMOID_HARD"
    elif activation == _keras.activations.tanh:
        activation_str = "TANH"
    elif activation == _keras.activations.relu:
        activation_str = "RELU"
    elif activation == _keras.activations.linear:
        activation_str = "LINEAR"
    else:
        raise NotImplementedError(
            "activation %s not supported for Recurrent layer." % activation
        )

    return activation_str


def _get_activation_name_from_keras_layer(keras_layer):
    if isinstance(keras_layer, _keras.layers.advanced_activations.LeakyReLU):
        non_linearity = "LEAKYRELU"
    elif isinstance(keras_layer, _keras.layers.advanced_activations.PReLU):
        non_linearity = "PRELU"
    elif isinstance(keras_layer, _keras.layers.advanced_activations.ELU):
        non_linearity = "ELU"
    elif isinstance(keras_layer, _keras.layers.advanced_activations.ThresholdedReLU):
        non_linearity = "THRESHOLDEDRELU"
    elif isinstance(keras_layer, _keras.layers.advanced_activations.Softmax):
        non_linearity = "SOFTMAX"
    else:
        act_name = keras_layer.activation.__name__

        if act_name == "softmax":
            non_linearity = "SOFTMAX"
        elif act_name == "sigmoid":
            non_linearity = "SIGMOID"
        elif act_name == "tanh":
            non_linearity = "TANH"
        elif act_name == "relu":
            non_linearity = "RELU"
        elif act_name == "relu6":
            non_linearity = "RELU6"
        elif act_name == "softplus":
            non_linearity = "SOFTPLUS"
        elif act_name == "softsign":
            non_linearity = "SOFTSIGN"
        elif act_name == "hard_sigmoid":
            non_linearity = "SIGMOID_HARD"
        elif act_name == "elu":
            non_linearity = "UNIT_ELU"
        elif act_name == "linear":
            non_linearity = "LINEAR"
        elif act_name == "selu":
            non_linearity = "SELU"
        else:
            non_linearity = "CUSTOM"

    return non_linearity


def _get_elementwise_name_from_keras_layer(keras_layer):
    """
    Get the keras layer name from the activation name.
    """
    if isinstance(keras_layer, _keras.layers.Add):
        return "ADD"
    elif isinstance(keras_layer, _keras.layers.Multiply):
        return "MULTIPLY"
    elif isinstance(keras_layer, _keras.layers.Concatenate):
        if len(keras_layer.input_shape[0]) == 3 and (
            keras_layer.axis == 1 or keras_layer.axis == -2
        ):
            return "SEQUENCE_CONCAT"
        if len(keras_layer.input_shape[0]) == 3 and (
            keras_layer.axis == 2 or keras_layer.axis == -1
        ):
            return "CONCAT"
        elif len(keras_layer.input_shape[0]) == 4 and (
            keras_layer.axis == 3 or keras_layer.axis == -1
        ):
            return "CONCAT"
        elif len(keras_layer.input_shape[0]) == 2 and (
            keras_layer.axis == 1 or keras_layer.axis == -1
        ):
            return "CONCAT"
        else:
            raise ValueError("Only channel and sequence concatenation are supported.")
    elif isinstance(keras_layer, _keras.layers.Dot):
        if len(keras_layer.input_shape[0]) == 2:
            if type(keras_layer.axes) is list or type(keras_layer.axes) is tuple:
                if len(keras_layer.axes) > 1:
                    raise ValueError("Only vector dot-product is supported.")
                else:
                    axis = keras_layer.axes[0]
            else:
                axis = keras_layer.axes
            if axis != -1 and axis != 1:
                raise ValueError("Only vector dot-product is supported.")

            if keras_layer.normalize:
                return "COS"
            else:
                return "DOT"
        else:
            raise ValueError("Only vector dot-product is supported.")
    elif isinstance(keras_layer, _keras.layers.Maximum):
        return "MAX"
    elif isinstance(keras_layer, _keras.layers.Average):
        return "AVE"
    else:
        _utils.raise_error_unsupported_option(
            str(type(keras_layer)), "merge", keras_layer.name
        )


def _same_elements_per_channel(x):
    """ Test if a 3D (H,W,C) matrix x has the same element in each (H,W)
    matrix for each channel
    """
    eps = 1e-5
    dims = x.shape
    for c in range(dims[-1]):
        xc = x[:, :, c].flatten()
        if not _np.all(_np.absolute(xc - xc[0]) < eps):
            return False
    return True


def _check_data_format(keras_layer):
    if hasattr(keras_layer, ("data_format")):
        if keras_layer.data_format != "channels_last":
            raise ValueError(
                "Converter currently supports data_format = " "'channels_last' only."
            )


def convert_dense(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert a dense layer from keras to coreml.

    Parameters
    keras_layer: layer
    ----------
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether or not to carry over Keras' "trainable" parameter.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    has_bias = keras_layer.use_bias
    # Get the weights from keras
    W = keras_layer.get_weights()[0].T
    Wb = keras_layer.get_weights()[1].T if has_bias else None
    output_channels, input_channels = W.shape

    builder.add_inner_product(
        name=layer,
        W=W,
        b=Wb,
        input_channels=input_channels,
        output_channels=output_channels,
        has_bias=has_bias,
        input_name=input_name,
        output_name=output_name,
    )
    if respect_train and keras_layer.trainable:
        builder.make_updatable([layer])


def convert_embedding(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """Convert a dense layer from keras to coreml.

    Parameters
    keras_layer: layer
    ----------
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether to support Keras' "trainable" flag (unsupported).
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    # Get the weights from keras
    W = keras_layer.get_weights()[0].T

    # assuming keras embedding layers don't have biases
    builder.add_embedding(
        name=layer,
        W=W,
        b=None,
        input_dim=keras_layer.input_dim,
        output_channels=keras_layer.output_dim,
        has_bias=False,
        input_name=input_name,
        output_name=output_name,
    )

    if respect_train and keras_layer.trainable:
        logging.warning(
            "Embedding layer '%s' is marked updatable, but Core "
            "ML does not yet support updating layers of this "
            "type. The layer will be frozen in Core ML.",
            layer,
        )


def convert_activation(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert an activation layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean,
        Ignored.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])
    non_linearity = _get_activation_name_from_keras_layer(keras_layer)

    # Add a non-linearity layer
    if non_linearity == "SOFTMAX":
        builder.add_softmax(name=layer, input_name=input_name, output_name=output_name)
        return
    if non_linearity == "RELU6":
        # No direct support of RELU with max-activation value - use negate and
        # clip layers
        relu_output_name = output_name + "_relu"
        builder.add_activation(layer, "RELU", input_name, relu_output_name)
        # negate it
        neg_output_name = relu_output_name + "_neg"
        builder.add_activation(
            layer + "__neg__", "LINEAR", relu_output_name, neg_output_name, [-1.0, 0]
        )
        # apply threshold
        clip_output_name = relu_output_name + "_clip"
        builder.add_unary(
            layer + "__clip__",
            neg_output_name,
            clip_output_name,
            "threshold",
            alpha=-6.0,
        )
        # negate it back
        builder.add_activation(
            layer + "_neg2", "LINEAR", clip_output_name, output_name, [-1.0, 0]
        )
        return

    if non_linearity == "SELU":
        elu_output_name = output_name + "_elu"
        builder.add_activation(
            layer + "__elu__", "ELU", input_name, elu_output_name, params=1.6732
        )
        builder.add_elementwise(
            layer,
            input_names=elu_output_name,
            output_name=output_name,
            mode="MULTIPLY",
            alpha=1.0507,
        )
        return

    params = None
    if non_linearity == "UNIT_ELU":
        params = 1.0
        non_linearity = "ELU"
    elif non_linearity == "LEAKYRELU":
        params = [keras_layer.alpha]
    elif non_linearity == "PRELU":
        shared_axes = list(keras_layer.shared_axes)
        if not (shared_axes == [1, 2, 3] or shared_axes == [1, 2]):
            _utils.raise_error_unsupported_scenario(
                "Shared axis not being [1,2,3] or [1,2]", "parametric_relu", layer
            )
        params = _keras.backend.eval(keras_layer.weights[0])
    elif non_linearity == "ELU":
        params = keras_layer.alpha
    elif non_linearity == "THRESHOLDEDRELU":
        params = keras_layer.theta
    else:
        pass  # do nothing to parameters

    builder.add_activation(
        name=layer,
        non_linearity=non_linearity,
        input_name=input_name,
        output_name=output_name,
        params=params,
    )


def convert_advanced_relu(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert an ReLU layer with maximum value from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Ignored.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    if keras_layer.max_value is None:
        builder.add_activation(layer, "RELU", input_name, output_name)
        return

    # No direct support of RELU with max-activation value - use negate and
    # clip layers
    relu_output_name = output_name + "_relu"
    builder.add_activation(layer, "RELU", input_name, relu_output_name)
    # negate it
    neg_output_name = relu_output_name + "_neg"
    builder.add_activation(
        layer + "__neg__", "LINEAR", relu_output_name, neg_output_name, [-1.0, 0]
    )
    # apply threshold
    clip_output_name = relu_output_name + "_clip"
    builder.add_unary(
        layer + "__clip__",
        neg_output_name,
        clip_output_name,
        "threshold",
        alpha=-keras_layer.max_value,
    )
    # negate it back
    builder.add_activation(
        layer + "_neg2", "LINEAR", clip_output_name, output_name, [-1.0, 0]
    )


def convert_convolution(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert convolution layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether or not to carry over Keras' "trainable" parameter.
    """

    _check_data_format(keras_layer)

    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    has_bias = keras_layer.use_bias
    is_deconv = isinstance(keras_layer, _keras.layers.convolutional.Conv2DTranspose)

    # Get the weights from _keras.
    weightList = keras_layer.get_weights()

    # Dimensions and weights
    if is_deconv:
        height, width, n_filters, channels = weightList[0].shape
        W = weightList[0].transpose([0, 1, 3, 2])
        try:
            output_blob_shape = list(filter(None, keras_layer.output_shape))
            output_shape = output_blob_shape[:-1]
        except:
            output_shape = None
    else:
        height, width, channels, n_filters = weightList[0].shape
        W = weightList[0]
        output_shape = None
    b = weightList[1] if has_bias else None

    output_channels = n_filters

    stride_height, stride_width = keras_layer.strides

    # Dilations
    dilations = [1, 1]
    if (type(keras_layer.dilation_rate) is list) or (
        type(keras_layer.dilation_rate) is tuple
    ):
        dilations = [keras_layer.dilation_rate[0], keras_layer.dilation_rate[1]]
    else:
        dilations = [keras_layer.dilation_rate, keras_layer.dilation_rate]
    if is_deconv and not dilations == [1, 1]:
        raise ValueError("Unsupported non-unity dilation for Deconvolution layer")

    groups = 1
    kernel_channels = channels
    # depth-wise convolution
    if isinstance(keras_layer, DepthwiseConv2D):
        groups = channels
        kernel_channels = 1
        depth_multiplier = keras_layer.depth_multiplier
        W = _np.reshape(W, (height, width, 1, channels * depth_multiplier))
        output_channels = channels * depth_multiplier

    builder.add_convolution(
        name=layer,
        kernel_channels=kernel_channels,
        output_channels=output_channels,
        height=height,
        width=width,
        stride_height=stride_height,
        stride_width=stride_width,
        border_mode=keras_layer.padding,
        groups=groups,
        W=W,
        b=b,
        has_bias=has_bias,
        is_deconv=is_deconv,
        output_shape=output_shape,
        input_name=input_name,
        output_name=output_name,
        dilation_factors=dilations,
    )

    if respect_train and keras_layer.trainable:
        builder.make_updatable([layer])


def convert_convolution1d(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert convolution layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether to honor Keras' "trainable" flag.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    has_bias = keras_layer.use_bias

    # Get the weights from _keras.
    # Keras stores convolution weights as list of numpy arrays
    weightList = keras_layer.get_weights()
    output_shape = list(filter(None, keras_layer.output_shape))[:-1]

    # Parameter
    filter_length, input_dim, n_filters = weightList[0].shape
    stride_width = (
        keras_layer.strides
        if type(keras_layer.strides) is int
        else keras_layer.strides[0]
    )

    # Weights and bias terms
    W = _np.expand_dims(weightList[0], axis=0)
    b = weightList[1] if has_bias else None

    dilations = [1, 1]
    if (type(keras_layer.dilation_rate) is list) or (
        type(keras_layer.dilation_rate) is tuple
    ):
        dilations = [1, keras_layer.dilation_rate[0]]
    else:
        dilations = [1, keras_layer.dilation_rate]

    keras_padding = keras_layer.padding
    if keras_padding == "causal":
        builder.add_padding(
            name=layer + "__causal_pad__",
            left=filter_length - 1,
            right=0,
            top=0,
            bottom=0,
            value=0,
            input_name=input_name,
            output_name=input_name + "__causal_pad__",
        )
        input_name = input_name + "__causal_pad__"
        keras_padding = "valid"

    builder.add_convolution(
        name=layer,
        kernel_channels=input_dim,
        output_channels=n_filters,
        height=1,
        width=filter_length,
        stride_height=1,
        stride_width=stride_width,
        border_mode=keras_padding,
        groups=1,
        W=W,
        b=b,
        has_bias=has_bias,
        is_deconv=False,
        output_shape=output_shape,
        input_name=input_name,
        output_name=output_name,
        dilation_factors=dilations,
    )

    if respect_train and keras_layer.trainable:
        builder.make_updatable([layer])


def convert_separable_convolution(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert separable convolution layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether to honor Keras' "trainable" flag.
    """
    _check_data_format(keras_layer)

    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    has_bias = keras_layer.use_bias

    # Get the weights from _keras.
    weight_list = keras_layer.get_weights()
    output_blob_shape = list(filter(None, keras_layer.output_shape))
    output_channels = output_blob_shape[-1]

    # D: depth mutliplier
    # w[0] is (H,W,Cin,D)
    # w[1] is (1,1,Cin * D, Cout)
    W0 = weight_list[0]
    W1 = weight_list[1]
    height, width, input_channels, depth_mult = W0.shape
    b = weight_list[2] if has_bias else None

    W0 = _np.reshape(W0, (height, width, 1, input_channels * depth_mult))

    stride_height, stride_width = keras_layer.strides

    # Dilations
    if (type(keras_layer.dilation_rate) is list) or (
        type(keras_layer.dilation_rate) is tuple
    ):
        dilations = [keras_layer.dilation_rate[0], keras_layer.dilation_rate[1]]
    else:
        dilations = [keras_layer.dilation_rate, keras_layer.dilation_rate]

    intermediate_name = output_name + "_intermin_"

    builder.add_convolution(
        name=layer + "_step_1",
        kernel_channels=1,
        output_channels=input_channels * depth_mult,
        height=height,
        width=width,
        stride_height=stride_height,
        stride_width=stride_width,
        border_mode=keras_layer.padding,
        groups=input_channels,
        W=W0,
        b=None,
        has_bias=False,
        is_deconv=False,
        output_shape=None,
        input_name=input_name,
        output_name=intermediate_name,
        dilation_factors=dilations,
    )

    builder.add_convolution(
        name=layer + "_step_2",
        kernel_channels=input_channels * depth_mult,
        output_channels=output_channels,
        height=1,
        width=1,
        stride_height=1,
        stride_width=1,
        border_mode=keras_layer.padding,
        groups=1,
        W=W1,
        b=b,
        has_bias=has_bias,
        is_deconv=False,
        output_shape=None,
        input_name=intermediate_name,
        output_name=output_name,
        dilation_factors=[1, 1],
    )

    if respect_train and keras_layer.trainable:
        builder.make_updatable([layer + "_step_1", layer + "_step_2"])


def convert_batchnorm(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert a Batch Normalization layer.

    Parameters
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether to honor Keras' "trainable" flag (unsupported).
    """

    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    axis = keras_layer.axis
    nb_channels = keras_layer.input_shape[axis]

    # Set parameters
    # Parameter arrangement in Keras: gamma, beta, mean, variance
    idx = 0
    gamma, beta = None, None
    if keras_layer.scale:
        gamma = keras_layer.get_weights()[idx]
        idx += 1
    if keras_layer.center:
        beta = keras_layer.get_weights()[idx]
        idx += 1
    mean = keras_layer.get_weights()[idx]
    std = keras_layer.get_weights()[idx + 1]

    gamma = _np.ones(mean.shape) if gamma is None else gamma
    beta = _np.zeros(mean.shape) if beta is None else beta

    # compute adjusted parameters
    variance = std * std
    f = 1.0 / _np.sqrt(std + keras_layer.epsilon)
    gamma1 = gamma * f
    beta1 = beta - gamma * mean * f
    mean[:] = 0.0  # mean
    variance[:] = 1.0 - 0.00001  # stddev

    builder.add_batchnorm(
        name=layer,
        channels=nb_channels,
        gamma=gamma1,
        beta=beta1,
        mean=mean,
        variance=variance,
        input_name=input_name,
        output_name=output_name,
    )

    if respect_train and keras_layer.trainable:
        logging.warning(
            "BatchNorm layer '%s' is marked updatable, but Core "
            "ML does not yet support updating layers of this "
            "type. The layer will be frozen in Core ML.",
            layer,
        )


def convert_flatten(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert a flatten layer from keras to coreml.
    ----------
    Parameters
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Ignored.
    """
    input_name, output_name = (input_names[0], output_names[0])

    # blob_order == 0 if the input blob needs not be rearranged
    # blob_order == 1 if the input blob needs to be rearranged
    blob_order = 0

    # using keras_layer.input.shape have a "?" (Dimension[None] at the front),
    # making a 3D tensor with unknown batch size 4D

    try:
        in_shape = keras_layer.input_shape
        if len(in_shape) == 4:
            blob_order = 1
        if len(in_shape) == 3 and in_shape[0] is None:
            # handling Keras rank-3 tensor (Batch, Sequence, Channels)
            permute_output_name = output_name + "__permute__"
            builder.add_permute(
                name=layer + "__permute__",
                dim=(2, 1, 0, 3),
                input_name=input_name,
                output_name=permute_output_name,
            )
            builder.add_flatten(
                name=layer,
                mode=1,
                input_name=permute_output_name,
                output_name=output_name,
            )
        else:
            builder.add_flatten(
                name=layer,
                mode=blob_order,
                input_name=input_name,
                output_name=output_name,
            )
    except:
        builder.add_flatten(
            name=layer, mode=1, input_name=input_name, output_name=output_name
        )


def convert_merge(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert concat layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Ignored.
    """
    # Get input and output names
    output_name = output_names[0]

    mode = _get_elementwise_name_from_keras_layer(keras_layer)
    builder.add_elementwise(
        name=layer, input_names=input_names, output_name=output_name, mode=mode
    )


def convert_pooling(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert pooling layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Ignored.
    """
    _check_data_format(keras_layer)
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    # Pooling layer type
    if (
        isinstance(keras_layer, _keras.layers.convolutional.MaxPooling2D)
        or isinstance(keras_layer, _keras.layers.convolutional.MaxPooling1D)
        or isinstance(keras_layer, _keras.layers.pooling.GlobalMaxPooling2D)
        or isinstance(keras_layer, _keras.layers.pooling.GlobalMaxPooling1D)
    ):
        layer_type_str = "MAX"
    elif (
        isinstance(keras_layer, _keras.layers.convolutional.AveragePooling2D)
        or isinstance(keras_layer, _keras.layers.convolutional.AveragePooling1D)
        or isinstance(keras_layer, _keras.layers.pooling.GlobalAveragePooling2D)
        or isinstance(keras_layer, _keras.layers.pooling.GlobalAveragePooling1D)
    ):
        layer_type_str = "AVERAGE"
    else:
        raise TypeError("Pooling type %s not supported" % keras_layer)

    # if it's global, set the global flag
    if isinstance(keras_layer, _keras.layers.pooling.GlobalMaxPooling2D) or isinstance(
        keras_layer, _keras.layers.pooling.GlobalAveragePooling2D
    ):
        # 2D global pooling
        global_pooling = True
        height, width = (0, 0)
        stride_height, stride_width = (0, 0)
        padding_type = "VALID"
    elif isinstance(
        keras_layer, _keras.layers.pooling.GlobalMaxPooling1D
    ) or isinstance(keras_layer, _keras.layers.pooling.GlobalAveragePooling1D):
        # 1D global pooling: 1D global pooling seems problematic in the backend,
        # use this work-around
        global_pooling = False
        _, width, channels = keras_layer.input_shape
        height = 1
        stride_height, stride_width = height, width
        padding_type = "VALID"
    else:
        global_pooling = False
        # Set pool sizes and strides
        # 1D cases:
        if (
            isinstance(keras_layer, _keras.layers.convolutional.MaxPooling1D)
            or isinstance(keras_layer, _keras.layers.pooling.GlobalMaxPooling1D)
            or isinstance(keras_layer, _keras.layers.convolutional.AveragePooling1D)
            or isinstance(keras_layer, _keras.layers.pooling.GlobalAveragePooling1D)
        ):
            pool_size = (
                keras_layer.pool_size
                if type(keras_layer.pool_size) is int
                else keras_layer.pool_size[0]
            )
            height, width = 1, pool_size
            if keras_layer.strides is not None:
                strides = (
                    keras_layer.strides
                    if type(keras_layer.strides) is int
                    else keras_layer.strides[0]
                )
                stride_height, stride_width = 1, strides
            else:
                stride_height, stride_width = 1, pool_size
        # 2D cases:
        else:
            height, width = keras_layer.pool_size
            if keras_layer.strides is None:
                stride_height, stride_width = height, width
            else:
                stride_height, stride_width = keras_layer.strides

        # Padding
        padding = keras_layer.padding
        if keras_layer.padding == "valid":
            padding_type = "VALID"
        elif keras_layer.padding == "same":
            padding_type = "SAME"
        else:
            raise TypeError("Border mode %s not supported" % padding)

    builder.add_pooling(
        name=layer,
        height=height,
        width=width,
        stride_height=stride_height,
        stride_width=stride_width,
        layer_type=layer_type_str,
        padding_type=padding_type,
        input_name=input_name,
        output_name=output_name,
        exclude_pad_area=True,
        is_global=global_pooling,
    )


def convert_padding(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert padding layer from keras to coreml.
    Keras only supports zero padding at this time.
    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Ignored.
    """
    _check_data_format(keras_layer)
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    is_1d = isinstance(keras_layer, _keras.layers.ZeroPadding1D)

    padding = keras_layer.padding
    top = left = bottom = right = 0
    if is_1d:
        if type(padding) is int:
            left = right = padding
        elif type(padding) is tuple:
            if type(padding[0]) is int:
                left, right = padding
            elif type(padding[0]) is tuple and len(padding[0]) == 2:
                left, right = padding[0]
            else:
                raise ValueError("Unrecognized padding option: %s" % (str(padding)))
        else:
            raise ValueError("Unrecognized padding option: %s" % (str(padding)))
    else:
        if type(padding) is int:
            top = left = bottom = right = padding
        elif type(padding) is tuple:
            if type(padding[0]) is int:
                top, left = padding
                bottom, right = padding
            elif type(padding[0]) is tuple:
                top, bottom = padding[0]
                left, right = padding[1]
            else:
                raise ValueError("Unrecognized padding option: %s" % (str(padding)))
        else:
            raise ValueError("Unrecognized padding option: %s" % (str(padding)))

    # Now add the layer
    builder.add_padding(
        name=layer,
        left=left,
        right=right,
        top=top,
        bottom=bottom,
        value=0,
        input_name=input_name,
        output_name=output_name,
    )


def convert_cropping(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert padding layer from keras to coreml.
    Keras only supports zero padding at this time.
    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Ignored.
    """

    _check_data_format(keras_layer)
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])
    is_1d = isinstance(keras_layer, _keras.layers.Cropping1D)

    cropping = keras_layer.cropping
    top = left = bottom = right = 0
    if is_1d:
        if type(cropping) is int:
            left = right = cropping
        elif type(cropping) is tuple:
            if type(cropping[0]) is int:
                left, right = cropping
            elif type(cropping[0]) is tuple and len(cropping[0]) == 2:
                left, right = cropping[0]
            else:
                raise ValueError("Unrecognized cropping option: %s" % (str(cropping)))
        else:
            raise ValueError("Unrecognized cropping option: %s" % (str(cropping)))
    else:
        if type(cropping) is int:
            top = left = bottom = right = cropping
        elif type(cropping) is tuple:
            if type(cropping[0]) is int:
                top, left = cropping
                bottom, right = cropping
            elif type(cropping[0]) is tuple:
                top, bottom = cropping[0]
                left, right = cropping[1]
            else:
                raise ValueError("Unrecognized cropping option: %s" % (str(cropping)))
        else:
            raise ValueError("Unrecognized cropping option: %s" % (str(cropping)))

    # Now add the layer
    builder.add_crop(
        name=layer,
        left=left,
        right=right,
        top=top,
        bottom=bottom,
        offset=[0, 0],
        input_names=[input_name],
        output_name=output_name,
    )


def convert_upsample(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert upsample layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Ignored.
    """
    _check_data_format(keras_layer)
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    is_1d = isinstance(keras_layer, _keras.layers.UpSampling1D)

    # Currently, we only support upsample of same dims
    fh = fw = 1
    if is_1d:
        if type(keras_layer.size) is tuple and len(keras_layer.size) == 1:
            fh, fw = 1, keras_layer.size[0]
        elif type(keras_layer.size) is int:
            fh, fw = 1, keras_layer.size
        else:
            raise ValueError(
                "Unrecognized upsample factor format %s" % (str(keras_layer.size))
            )
    else:
        if type(keras_layer.size) is int:
            fh = fw = keras_layer.size
        elif len(keras_layer.size) == 2:
            if keras_layer.size[0] != keras_layer.size[1]:
                raise ValueError(
                    "Upsample with different rows and columns not " "supported."
                )
            else:
                fh = keras_layer.size[0]
                fw = keras_layer.size[1]
        else:
            raise ValueError(
                "Unrecognized upsample factor format %s" % (str(keras_layer.size))
            )

    kerasmode2coreml = {"nearest": "NN", "bilinear": "BILINEAR"}
    interpolation = getattr(
        keras_layer, "interpolation", "nearest"
    )  # Defaults to 'nearest' for Keras < 2.2.3

    if interpolation not in kerasmode2coreml:
        raise ValueError(
            'Only supported "nearest" or "bilinear" interpolation for upsampling layers.'
        )

    mode = kerasmode2coreml[interpolation]

    builder.add_upsample(
        name=layer,
        scaling_factor_h=fh,
        scaling_factor_w=fw,
        input_name=input_name,
        output_name=output_name,
        mode=mode,
    )


def convert_permute(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert a softmax layer from keras to coreml.

    Parameters
    keras_layer: layer
    ----------
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Ignored.
    """
    input_name, output_name = (input_names[0], output_names[0])

    keras_dims = keras_layer.dims
    # Keras permute layer index begins at 1
    if len(keras_dims) == 3:
        # Keras input tensor interpret as (H,W,C)
        x = list(_np.array(keras_dims))
        arr = [2, 3, 1]  # HWC in Keras
        arr_permuted = [arr[x[0] - 1], arr[x[1] - 1], arr[x[2] - 1]]
        arr_permuted = [
            arr_permuted[2],
            arr_permuted[0],
            arr_permuted[1],
        ]  # coreml format: channel first
        # add a sequence axis
        dim = [0] + arr_permuted
        dim = tuple(dim)
    elif len(keras_dims) == 4:
        # Here we use Keras converter as a place holder for inserting
        # permutations - the values here are not valid Keras dim parameters
        # but parameters we need to use to convert to CoreML model
        dim = keras_dims
    else:
        raise NotImplementedError("Supports only 3d permutation.")

    builder.add_permute(
        name=layer, dim=dim, input_name=input_name, output_name=output_name
    )


def convert_reshape(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    respect_train: boolean
        Ignored.
    """

    input_name, output_name = (input_names[0], output_names[0])

    input_shape = keras_layer.input_shape
    target_shape = keras_layer.target_shape

    def get_coreml_target_shape(target_shape):
        if len(target_shape) == 1:  # (D,)
            coreml_shape = (1, target_shape[0], 1, 1)
        elif len(target_shape) == 2:  # (S,D)
            coreml_shape = target_shape + (1, 1)
        elif len(target_shape) == 3:  # (H,W,C)
            coreml_shape = (1, target_shape[2], target_shape[0], target_shape[1])
        else:
            coreml_shape = None
        return coreml_shape

    def get_mode(input_shape, target_shape):
        in_shape = input_shape[1:]
        if len(in_shape) == 3 or len(target_shape) == 3:
            return 1
        else:
            return 0

    new_shape = get_coreml_target_shape(target_shape)
    if new_shape is not None:
        mode = get_mode(input_shape, target_shape)
        builder.add_reshape(
            name=layer,
            input_name=input_name,
            output_name=output_name,
            target_shape=new_shape,
            mode=mode,
        )
    else:
        _utils.raise_error_unsupported_categorical_option(
            "input_shape", str(input_shape), "reshape", layer
        )


def convert_simple_rnn(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert an SimpleRNN layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether to honor Keras' "trainable" flag (unsupported).
    """
    # Get input and output names
    hidden_size = keras_layer.units
    input_size = keras_layer.input_shape[-1]

    output_all = keras_layer.return_sequences
    reverse_input = keras_layer.go_backwards

    W_h = _np.zeros((hidden_size, hidden_size))
    W_x = _np.zeros((hidden_size, input_size))
    b = None

    implementation = (
        keras_layer.implementation if hasattr(keras_layer, "implementation") else 0
    )
    if implementation == 0:
        W_h = keras_layer.get_weights()[1].T
        W_x = keras_layer.get_weights()[0].T
        if keras_layer.use_bias:
            b = keras_layer.get_weights()[2]

    # Set actication type
    activation_str = _get_recurrent_activation_name_from_keras(keras_layer.activation)

    # Add to the network
    builder.add_simple_rnn(
        name=layer,
        W_h=W_h,
        W_x=W_x,
        b=b,
        hidden_size=hidden_size,
        input_size=input_size,
        activation=activation_str,
        input_names=input_names,
        output_names=output_names,
        output_all=output_all,
        reverse_input=reverse_input,
    )

    if respect_train and keras_layer.trainable:
        logging.warning(
            "RNN layer '%s' is marked updatable, but Core "
            "ML does not yet support updating layers of this "
            "type. The layer will be frozen in Core ML.",
            layer,
        )


def convert_lstm(builder, layer, input_names, output_names, keras_layer, respect_train):
    """
    Convert an LSTM layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether to honor Keras' "trainable" flag (unsupported).
    """

    hidden_size = keras_layer.units
    input_size = keras_layer.input_shape[-1]
    output_all = keras_layer.return_sequences
    reverse_input = keras_layer.go_backwards

    # Keras: [W_x, W_h, b] each in I F C O
    # CoreML: I F O G; W_h and W_x are separated
    W_h, W_x, b = ([], [], [])
    keras_W_h = keras_layer.get_weights()[1].T
    W_h.append(keras_W_h[0 * hidden_size :][:hidden_size])
    W_h.append(keras_W_h[1 * hidden_size :][:hidden_size])
    W_h.append(keras_W_h[3 * hidden_size :][:hidden_size])
    W_h.append(keras_W_h[2 * hidden_size :][:hidden_size])

    keras_W_x = keras_layer.get_weights()[0].T
    W_x.append(keras_W_x[0 * hidden_size :][:hidden_size])
    W_x.append(keras_W_x[1 * hidden_size :][:hidden_size])
    W_x.append(keras_W_x[3 * hidden_size :][:hidden_size])
    W_x.append(keras_W_x[2 * hidden_size :][:hidden_size])
    if keras_layer.use_bias:
        keras_b = keras_layer.get_weights()[2]
        b.append(keras_b[0 * hidden_size :][:hidden_size])
        b.append(keras_b[1 * hidden_size :][:hidden_size])
        b.append(keras_b[3 * hidden_size :][:hidden_size])
        b.append(keras_b[2 * hidden_size :][:hidden_size])
    if len(b) == 0:
        b = None

    # Set activation type
    inner_activation_str = _get_recurrent_activation_name_from_keras(
        keras_layer.recurrent_activation
    )
    activation_str = _get_recurrent_activation_name_from_keras(keras_layer.activation)

    # Add to the network
    builder.add_unilstm(
        name=layer,
        W_h=W_h,
        W_x=W_x,
        b=b,
        hidden_size=hidden_size,
        input_size=input_size,
        input_names=input_names,
        output_names=output_names,
        inner_activation=inner_activation_str,
        cell_state_update_activation=activation_str,
        output_activation=activation_str,
        output_all=output_all,
        forget_bias=keras_layer.unit_forget_bias,
        reverse_input=reverse_input,
    )

    if respect_train and keras_layer.trainable:
        logging.warning(
            "LSTM layer '%s' is marked updatable, but Core "
            "ML does not yet support updating layers of this "
            "type. The layer will be frozen in Core ML.",
            layer,
        )


def convert_gru(builder, layer, input_names, output_names, keras_layer, respect_train):
    """
    Convert a GRU layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether to honor Keras' "trainable" flag (unsupported).
    """

    hidden_size = keras_layer.units
    input_size = keras_layer.input_shape[-1]

    output_all = keras_layer.return_sequences
    reverse_input = keras_layer.go_backwards

    # Keras: Z R O
    # CoreML: Z R O
    W_h, W_x, b = ([], [], [])
    keras_W_h = keras_layer.get_weights()[1].T
    W_h.append(keras_W_h[0 * hidden_size :][:hidden_size])
    W_h.append(keras_W_h[1 * hidden_size :][:hidden_size])
    W_h.append(keras_W_h[2 * hidden_size :][:hidden_size])

    keras_W_x = keras_layer.get_weights()[0].T
    W_x.append(keras_W_x[0 * hidden_size :][:hidden_size])
    W_x.append(keras_W_x[1 * hidden_size :][:hidden_size])
    W_x.append(keras_W_x[2 * hidden_size :][:hidden_size])

    if keras_layer.use_bias:
        keras_b = keras_layer.get_weights()[2]
        b.append(keras_b[0 * hidden_size :][:hidden_size])
        b.append(keras_b[1 * hidden_size :][:hidden_size])
        b.append(keras_b[2 * hidden_size :][:hidden_size])
    if len(b) == 0:
        b = None

    # Set actication type
    inner_activation_str = _get_recurrent_activation_name_from_keras(
        keras_layer.recurrent_activation
    )
    activation_str = _get_recurrent_activation_name_from_keras(keras_layer.activation)

    # Add to the network
    builder.add_gru(
        name=layer,
        W_h=W_h,
        W_x=W_x,
        b=b,
        input_size=input_size,
        hidden_size=hidden_size,
        input_names=input_names,
        output_names=output_names,
        activation=activation_str,
        inner_activation=inner_activation_str,
        output_all=output_all,
        reverse_input=reverse_input,
    )

    if respect_train and keras_layer.trainable:
        logging.warning(
            "GRU layer '%s' is marked updatable, but Core "
            "ML does not yet support updating layers of this "
            "type. The layer will be frozen in Core ML.",
            layer,
        )


def convert_bidirectional(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    Convert a bidirectional layer from keras to coreml.
    Currently assumes the units are LSTMs.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    respect_train: boolean
        Whether to honor Keras' "trainable" flag (unsupported).
    """

    input_size = keras_layer.input_shape[-1]

    lstm_layer = keras_layer.forward_layer
    if type(lstm_layer) != _keras.layers.recurrent.LSTM:
        raise TypeError("Bidirectional layers only supported with LSTM")

    if lstm_layer.go_backwards:
        raise TypeError(" 'go_backwards' mode not supported with Bidirectional layers")

    output_all = keras_layer.return_sequences
    hidden_size = lstm_layer.units

    # Keras: I C F O; W_x, W_h, b
    # CoreML: I F O G; W_h and W_x are separated
    # Keras has all forward weights, followed by backward in the same order
    W_h, W_x, b = ([], [], [])
    keras_W_h = keras_layer.forward_layer.get_weights()[1].T
    W_h.append(keras_W_h[0 * hidden_size :][:hidden_size])
    W_h.append(keras_W_h[1 * hidden_size :][:hidden_size])
    W_h.append(keras_W_h[3 * hidden_size :][:hidden_size])
    W_h.append(keras_W_h[2 * hidden_size :][:hidden_size])

    keras_W_x = keras_layer.forward_layer.get_weights()[0].T
    W_x.append(keras_W_x[0 * hidden_size :][:hidden_size])
    W_x.append(keras_W_x[1 * hidden_size :][:hidden_size])
    W_x.append(keras_W_x[3 * hidden_size :][:hidden_size])
    W_x.append(keras_W_x[2 * hidden_size :][:hidden_size])

    if keras_layer.forward_layer.use_bias:
        keras_b = keras_layer.forward_layer.get_weights()[2]
        b.append(keras_b[0 * hidden_size :][:hidden_size])
        b.append(keras_b[1 * hidden_size :][:hidden_size])
        b.append(keras_b[3 * hidden_size :][:hidden_size])
        b.append(keras_b[2 * hidden_size :][:hidden_size])

    if len(b) == 0:
        b = None

    W_h_back, W_x_back, b_back = ([], [], [])
    keras_W_h = keras_layer.backward_layer.get_weights()[1].T
    W_h_back.append(keras_W_h[0 * hidden_size :][:hidden_size])
    W_h_back.append(keras_W_h[1 * hidden_size :][:hidden_size])
    W_h_back.append(keras_W_h[3 * hidden_size :][:hidden_size])
    W_h_back.append(keras_W_h[2 * hidden_size :][:hidden_size])

    keras_W_x = keras_layer.backward_layer.get_weights()[0].T
    W_x_back.append(keras_W_x[0 * hidden_size :][:hidden_size])
    W_x_back.append(keras_W_x[1 * hidden_size :][:hidden_size])
    W_x_back.append(keras_W_x[3 * hidden_size :][:hidden_size])
    W_x_back.append(keras_W_x[2 * hidden_size :][:hidden_size])

    if keras_layer.backward_layer.use_bias:
        keras_b = keras_layer.backward_layer.get_weights()[2]
        b_back.append(keras_b[0 * hidden_size :][:hidden_size])
        b_back.append(keras_b[1 * hidden_size :][:hidden_size])
        b_back.append(keras_b[3 * hidden_size :][:hidden_size])
        b_back.append(keras_b[2 * hidden_size :][:hidden_size])
    if len(b_back) == 0:
        b_back = None

    if (b == None and b_back != None) or (b != None and b_back == None):
        raise ValueError(
            "Unsupported Bi-directional LSTM configuration. Bias "
            "must be enabled/disabled for both directions."
        )

    # Set activation type
    inner_activation_str = _get_recurrent_activation_name_from_keras(
        lstm_layer.recurrent_activation
    )
    activation_str = _get_recurrent_activation_name_from_keras(lstm_layer.activation)

    output_name_1 = output_names[0]
    if hasattr(keras_layer, "merge_mode"):
        merge_mode = keras_layer.merge_mode
        if merge_mode not in ["concat", "sum", "mul", "ave"]:
            raise NotImplementedError(
                "merge_mode '%s' in Bidirectional LSTM "
                "not supported currently" % merge_mode
            )
        if merge_mode != "concat":
            output_name_1 += "_concatenated_bilstm_output"

    # Add to the network
    builder.add_bidirlstm(
        name=layer,
        W_h=W_h,
        W_x=W_x,
        b=b,
        W_h_back=W_h_back,
        W_x_back=W_x_back,
        b_back=b_back,
        hidden_size=hidden_size,
        input_size=input_size,
        input_names=input_names,
        output_names=[output_name_1] + output_names[1:],
        inner_activation=inner_activation_str,
        cell_state_update_activation=activation_str,
        output_activation=activation_str,
        forget_bias=lstm_layer.unit_forget_bias,
        output_all=output_all,
    )

    if output_name_1 != output_names[0]:
        mode = "CONCAT"
        if merge_mode == "sum":
            mode = "ADD"
        elif merge_mode == "ave":
            mode = "AVE"
        elif merge_mode == "mul":
            mode = "MULTIPLY"
        builder.add_split(
            name=layer + "_split",
            input_name=output_name_1,
            output_names=[output_names[0] + "_forward", output_names[0] + "_backward"],
        )
        builder.add_elementwise(
            name=layer + "_elementwise",
            input_names=[output_names[0] + "_forward", output_names[0] + "_backward"],
            output_name=output_names[0],
            mode=mode,
        )

    if respect_train and keras_layer.trainable:
        logging.warning(
            "Bidirectional layer '%s' is marked updatable, but "
            "Core ML does not yet support updating layers of this "
            "type. The layer will be frozen in Core ML.",
            layer,
        )


def convert_repeat_vector(
    builder, layer, input_names, output_names, keras_layer, respect_train
):
    """
    respect_train: boolean
        Ignored.
    """
    # Keras RepeatVector only deals with 1D input
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    builder.add_sequence_repeat(
        name=layer, nrep=keras_layer.n, input_name=input_name, output_name=output_name
    )


def default_skip(builder, layer, input_names, output_names, keras_layer, respect_train):
    """
    Layers that can be skipped.
    """
    return



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import scipy
from ._op_reqs import *

"""
Elementwise Unary Op Superclass
"""


class elementwise_unary(Operation):
    input_spec = InputSpec(x=ScalarOrTensorInputType(),)

    def __init__(self, **kwargs):
        super(elementwise_unary, self).__init__(**kwargs)

    def type_inference(self):
        return self.x.sym_type


"""
Elementwise unary op implementation(s)
"""


@register_op(doc_str="")
class abs(elementwise_unary):
    """
    Return the absolute values of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(abs, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.abs(self.x.val)


@register_op(doc_str="")
class acos(elementwise_unary):
    """
    Return the inverse cosine values of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(acos, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.arccos(self.x.val)


@register_op(doc_str="")
class asin(elementwise_unary):
    """
    Return the inverse sine of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(asin, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.arcsin(self.x.val)


@register_op(doc_str="")
class atan(elementwise_unary):
    """
    Return the inverse tangent of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(atan, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.arctan(self.x.val)


@register_op(doc_str="")
class atanh(elementwise_unary):
    """
    Return the inverse hyperbolic tangent values of the input
    ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(atanh, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.arctanh(self.x.val)


@register_op(doc_str="")
class ceil(elementwise_unary):
    """
    Return the ceil values of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(ceil, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.ceil(self.x.val)


@register_op(doc_str="")
class clip(Operation):
    """
    Clip the values in the input ``x`` to ``[alpha, beta]``, element-wise.
    Any values less than ``alpha`` are set to ``alpha``, and any values greater
    than ``beta`` are set to ``beta``.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    alpha: const f32 (Required)
    beta: const f32 (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    input_spec = InputSpec(
        x=ScalarOrTensorInputType(),
        alpha=FloatInputType(const=True),
        beta=FloatInputType(const=True),
    )

    def __init__(self, **kwargs):
        super(clip, self).__init__(**kwargs)

    def type_inference(self):
        return self.x.sym_type

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.minimum(np.maximum(self.x.val, self.alpha.val), self.beta.val)


@register_op(doc_str="")
class cos(elementwise_unary):
    """
    Return cosine of ``x`` element-wise. Input domain is ``(-inf, inf)`` and
    output range is ``[-1,1]``.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], T>
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(cos, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.cos(self.x.val)


@register_op(doc_str="")
class cosh(elementwise_unary):
    """
    Return hyperbolic cosine of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], T>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(cosh, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.cosh(self.x.val)


@register_op(doc_str="")
class erf(elementwise_unary):
    """
    Return the gauss error function of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(erf, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return scipy.special.erf(self.x.val)


@register_op(doc_str="")
class exp(elementwise_unary):
    """
    Return the exponential values of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(exp, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.exp(self.x.val)


@register_op(doc_str="")
class exp2(elementwise_unary):
    """
    Return the exponential values of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(exp2, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.exp2(self.x.val)


@register_op(doc_str="")
class floor(elementwise_unary):
    """
    Return the floor of the input ``x``, element-wise, the same as rounding
    towards negative infinity.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(floor, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.floor(self.x.val)


@register_op(doc_str="")
class inverse(Operation):
    """
    Return the reciprocal value of the input ``x``, element-wise.

    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    epsilon: const fp32 (Optional, default=1e-4)
        * This is a small constant that is added to the input, before taking its
          inverse, for stability.
        * ``y = 1 / (x + epsilon)``.

    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.

    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        x=ScalarOrTensorInputType(),
        epsilon=FloatInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            epsilon=1e-4,
            )

    def __init__(self, **kwargs):
        super(inverse, self).__init__(**kwargs)

    def type_inference(self):
        return self.x.sym_type

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.reciprocal(self.x.val + self.epsilon.val)


@register_op(doc_str="")
class log(Operation):
    """
    Return the natural logarithm value of the input ``x``, element-wise.

    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    epsilon: const fp32 (Optional, default=1e-45)
        * This is a small constant that is added to the input, before taking log.
        * ``y = log(x + epsilon)``.

    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.

    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        x=ScalarOrTensorInputType(),
        epsilon=FloatInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            epsilon=1e-45)

    def __init__(self, **kwargs):
        super(log, self).__init__(**kwargs)

    def type_inference(self):
        return self.x.sym_type

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.log(self.x.val + self.epsilon.val)


@register_op(doc_str="")
class logical_not(elementwise_unary):
    """
    Return the value of NOT the input ``x``, element-wise. (``1`` for true, ``0``
    for false in numeric domain.) A numeric value ``t`` is evaluated to true
    ``iff t != 0``.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(logical_not, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.logical_not(self.x.val)


@register_op(doc_str="")
class round(elementwise_unary):
    """
    Return the round value of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(round, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.round(self.x.val)


@register_op(doc_str="")
class rsqrt(Operation):
    """
    Return the reciprocal value of the square root of the input ``x``, element-wise.

    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    epsilon: const fp32 (Optional, default=1e-12)
        * This is a small constant that is added to the input, before applying the
          ``rsqrt`` function, for stability.
        * ``y = 1 / sqrt(x + epsilon)``.

    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.

    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        x=ScalarOrTensorInputType(),
        epsilon=FloatInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            epsilon=1e-12,
            )

    def __init__(self, **kwargs):
        super(rsqrt, self).__init__(**kwargs)

    def type_inference(self):
        return self.x.sym_type

    @precondition(allow=VALUE)
    def value_inference(self):
        return 1.0 / np.sqrt(self.x.val + self.epsilon.val)


@register_op(doc_str="")
class sign(elementwise_unary):
    """
    Return the sign value of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(sign, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.sign(self.x.val)


@register_op(doc_str="")
class sin(elementwise_unary):
    """
    Return the sine value of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(sin, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.sin(self.x.val)


@register_op(doc_str="")
class sinh(elementwise_unary):
    """
    Return the hyperbolic sine value of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(sinh, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.sinh(self.x.val)


@register_op(doc_str="")
class sqrt(elementwise_unary):
    """
    Returns the square root value of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(sqrt, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.sqrt(self.x.val)


@register_op(doc_str="")
class square(elementwise_unary):
    """
    Return the square value of the input ``x``, element-wise.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(square, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.square(self.x.val)


@register_op(doc_str="")
class tan(elementwise_unary):
    """
    Return the tangent value of the input ``x``, element-wise. Both input and output
    ranges are ``(-inf, inf)``.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(tan, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.tan(self.x.val)


@register_op(doc_str="")
class tanh(elementwise_unary):
    """
    Return the hyperbolic tangent value of the input ``x``, element-wise. Both input
    and output ranges are ``(-inf, inf)`` while output range is ``[-1,  1]``.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    def __init__(self, **kwargs):
        super(tanh, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.tanh(self.x.val)


@register_op(doc_str="")
class threshold(Operation):
    """
    Set a lower bound ``alpha`` to the values in the input ``x``, element-wise.
    Any values less than ``alpha`` are set to ``alpha``.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    alpha: const fp32 (Required)
    
    Returns
    -------
    tensor<[\*d], f32>
        * A tensor of the same shape as ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    
    input_spec = InputSpec(
        x=ScalarOrTensorInputType(), alpha=FloatInputType(const=True),
    )

    def __init__(self, **kwargs):
        super(threshold, self).__init__(**kwargs)

    def type_inference(self):
        return self.x.sym_type

    @precondition(allow=VALUE)
    def value_inference(self):
        return np.maximum(self.x.val, self.alpha.val)


@register_op(doc_str="")
class cast(Operation):
    """
    Cast the input ``x`` to the new type ``dtype``.
    
    Parameters
    ----------
    x: tensor<[\*d], T> (Required)
    dtype: const str (Required)
        * Can be one of the following types: ``int32``, ``int64``, ``fp32``, ``fp64``.
    
    Returns
    -------
    tensor<[\*d], dtype>
        * A tensor of the same shape as ``x``, with type ``dtype``.
    
    Attributes
    ----------
    T: i32, i64, fp32, fp64, bool.
    """
    
    input_spec = InputSpec(
        x=ScalarOrTensorInputType(), dtype=StringInputType(const=True)
    )

    def __init__(self, **kwargs):
        super(cast, self).__init__(**kwargs)

    def type_inference(self):
        type_map = {
            "int32": types.int32,
            "int64": types.int64,
            "fp32": types.fp32,
            "fp64": types.fp64,
            "bool": types.bool,
        }

        if self.dtype.val not in type_map.keys():
            raise NotImplementedError(
                "Parameter dtype of the cast operation can be one of the {}. "
                "Provided {}".format(type_map.keys(), self.dtype.val)
            )

        if not types.is_tensor(self.x.sym_type):
            return type_map[self.dtype.val]

        ret_shape = self.x.shape
        return types.tensor(type_map[self.dtype.val], ret_shape)

    @precondition(allow=VALUE)
    def value_inference(self):
        type_map = {
            "int32": np.int32,
            "int64": np.int64,
            "fp32": np.float32,
            "fp64": np.float64,
            "bool": np.bool,
        }

        if self.dtype.val not in type_map.keys():
            raise NotImplementedError(
                "Parameter dtype of the cast operation can be one of the {}. "
                "Provided {}".format(type_map.keys(), self.dtype.val)
            )

        if not types.is_tensor(self.x.sym_type):
            return self.x.val.astype(dtype=type_map[self.dtype.val])
        else:
            return np.array(self.x.val).astype(dtype=type_map[self.dtype.val])



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.frontend.tensorflow.converter import TFConverter
from coremltools.converters.mil.frontend.tensorflow.basic_graph_ops import (
    simple_topsort,
)

from .ssa_passes.tf_passes import tensorflow_passes as tensorflow2_passes


class TF2Converter(TFConverter):
    def __init__(self, tf_ssa, inputs=None, outputs=None, **kwargs):
        TFConverter.__init__(self, tf_ssa, inputs, outputs, **kwargs)

        # Overwrite tensorflow_passes
        # TF 2.x uses different set of graph passes
        self.tensorflow_passes = tensorflow2_passes

    def _get_stack(self, tfssa, root="main"):
        """
        Overwrite TFConverter._get_stack() as TF2 generates different sub-graphs.
        """

        # We're trying to get a order of how to loop through the graphs.
        # This is NOT necessarily a DAG.
        dep = {x: [] for x in tfssa.functions}
        for fname in tfssa.functions:
            for node in tfssa.functions[fname].graph.values():
                func_x, func_y = None, None

                if node.op in {"StatelessIf", "If"}:
                    func_x = node.attr.get("then_branch")
                    func_y = node.attr.get("else_branch")
                elif node.op in {"StatelessWhile", "While"}:
                    func_x = node.attr.get("body")
                    func_y = node.attr.get("cond")

                if func_x and fname not in dep[func_x]:
                    dep[func_x].append(fname)
                if func_y and fname not in dep[func_y]:
                    dep[func_y].append(fname)

        assert len(dep[root]) == 0
        graph_stack = simple_topsort(dep)

        return graph_stack



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.mil.var import Var
import logging
import numpy as np
import copy
from collections import defaultdict

DEBUG = False  # set to true to plot the block before and after the transformation

"""
Description of the Pass:

The pass is divided into 3 phases

1st phase: information gathering
---------------------------------

- We first plug in Identity ops for all output nodes. This allows us to treat all ops uniformly during traversal.
- Block is traversed in the topological order, starting from the ops connected to the inputs.
- During the traversal, a value is associated with every var in the block
- This value can be either of type "HypotheticalValue" or "LazyTransposeHypotheticalValue"
- Main purpose of type "HypotheticalValue" is to essentially indicate that it is NOT of type  "LazyTransposeHypotheticalValue"
- "LazyTransposeHypotheticalValue" represents either one or multiple transpose ops with the same perm value. This information
is stored in this class. It also wraps a "HypotheticalValue" that was the last hypothetical value which was generated
prior to the origin of "LazyTransposeHypotheticalValue"

- Each op decides which type of hypothetical value to associate with its output vars, based on its op type,
its attributes, and the types of the hypothetical values of its input vars
- Ops are classified into 4 categories: unary like, axis update, transpose and materialize (all the rest)
- Transpose ops: these are the ops from which a "LazyTransposeHypotheticalValue" originate.
    - If the input to it is a "HypotheticalValue", its output will be a "LazyTransposeHypotheticalValue",
        indicating that this transpose op is available to get cancelled downstream
    - If the input to it is a "LazyTransposeHypotheticalValue", then it is checked whether this op cancels it or not
    - If it cancels it, a "HypotheticalValue" value is generated at the output and the information about this transpose cancellation
      is recorded in the dictionary "transpose_op_to_cancel_ops"
    - If it does not cancel, the current transpose op is categrorized as a materialize op and hence the information in
       dictionary "transpose_op_to_materialize_ops" is updated accordingly. The output of the op is now mapped to a
       "HypotheticalValue"
- Unary like ops: they simply transfer their input hypothetical value type to the output.
- Axis update ops: if a transpose can pass through them, they are treated like a unary op and the dictionary
   "transpose_op_to_axis_update_ops" is updated. If the op cannot be updated in any manner to
   allow a transpose to pass through, this op is then categorized as a materialize op and handled accordingly
- Materialzie ops: All "LazyTransposeHypotheticalValue" input vars, if present, materialize here. Output of this op
  is always of type "HypotheticalValue". If the input is a "LazyTransposeHypotheticalValue", update the dict
  "transpose_op_to_materialize_ops"
- To treat an op like a unary op, add its type to "UNARY_LIKE_OP_TYPES". In future changes we want to make this process
automatic, by automatically detecting an op as a unary like by its "traits"

- To treat an op like axis update op, add a class specific to the op implementing the class "transform_axis_update_ops"
For examples, see classes "transform_concat", "transform_pad" etc. The dictionary AXIS_UPDATE_OPS is automatically filled
in by the decorator "register_axis_update_op"


2nd phase: making a determination on which transpose ops to remove from the graph
---------------------------------------------------------------------------------
- All Transpose ops that have a corresponding compliment op in dict "transpose_op_to_cancel_ops" is a candidate
- However, we need to make sure of two things
    - if a transpose op is removed then all its cancel ops in "transpose_op_to_cancel_ops" must be also removed,
      to ensure correctness of the graph. Same is true in the reverse direction as well,
      that is, for every cancel op that is removed all its parent transpose ops upstream, must also be removed.
    - transpose ops should only be removed if the number of cancel ops is greater than the number of transpose ops
      that would get freshly introduced to the block as a result of materialization ops. Right now in the algorithm
      each materialization op/output var (dicts "transpose_op_to_materialize_ops"/"old_output_vars"),
      results in one more transpose op, although this can be further optimized in the future

- To resolve this, we recognize that nodes consisting of sets (a) and (b) form a bipartitle graph, where,
(a) == starting transpose ops (originators of "LazyTransposeHypotheticalValue")
and (b) == set of transpose cancel ops and materialize ops
- in this bipartite graph, we find all the connected components
- for each connected component, either the whole set of transpose ops in it are removed/materialized or none
of them are touched
- thus for each set, a determination is made based on counting the number of cancel ops and materialize ops
- Based on this the final set of transpose ops to be removed is updated


3rd phase: transforming the graph
----------------------------------
- Transpose starting ops and the cancel ops are removed
- Axis update ops, affected by these transpose ops, are updated
- Transposes are materialized, i.e. added just before the materialize ops, which are linked to the starting transpose ops.
  The starting transpose op can get materialized (inserted) multiple times, before each of the "materialize ops" downstream.
- Block outputs are handled similar to the materialize ops
- Type inference on all ops is invoked after all the transformations
- All Identity ops that are plugged into the graph to treat outputs as materialized are removed.

Debugging:
------------
If the debug flag is set to True, the block before and after the transformation is plotted,
with transpose nodes highlighted

"""

# TODO: instead of a hard-coded set, use op-traits
# These are the ops that satisfy the following property:
# - single non constant input
# - single output
# - non rank changing
# - doesn't need to be updated of a transpose passes through it. i.e.
#  Transpose(op(x)) == op(Transpose(x))
UNARY_LIKE_OP_TYPES = set(
    [
        "relu",
        "log",
        "relu6",
        "abs",
        "acos",
        "asin",
        "atan",
        "atanh",
        "ceil",
        "clip",
        "cos",
        "cosh",
        "erf",
        "exp",
        "exp2",
        "floor",
        "identity",
        "logical_not",
        "round",
        "rsqrt",
        "sign",
        "sin",
        "sinh",
        "sqrt",
        "square",
        "tan",
        "tanh",
        "threshold",
        "clamped_relu",
        "elu",
        "gelu",
        "leaky_relu",
        "linear_activation",
        "scaled_tanh",
        "sigmoid",
        "sigmoid_hard",
        "softplus",
        "softplus_parametric",
        "softsign",
        "thresholded_relu",
    ]
)

# Dictionary from axis update op to its class
# This is filled in by child classes of the class "transform_axis_update_ops".
AXIS_UPDATE_OPS = {}


def _do_transposes_cancel(perm1, perm2):
    if len(perm1) != len(perm2):
        return False
    x = list(range(len(perm1)))
    x1 = [x[i] for i in perm1]
    x2 = [x1[i] for i in perm2]
    if x == x2:
        return True
    return False


def _get_input_vars(op, only_nonconst_vars=False):
    """
    :return: List[Var]
    """
    input_vars = []
    for name, val in op.inputs.items():
        if isinstance(val, Var):
            if only_nonconst_vars:
                if val.op and val.op.op_type == "const":
                    continue
            input_vars.append(val)
        elif isinstance(val, (list, tuple)):
            for var in val:
                if not isinstance(var, Var):
                    msg = "transpose optimization pass: unrecognized input type of op='{}', input='{}'"
                    raise ValueError(msg.format(op.name, name))
                if only_nonconst_vars:
                    if var.op and var.op.op_type == "const":
                        continue
                input_vars.append(var)
        else:
            msg = "transpose optimization pass: unrecognized input type of op='{}', input='{}'"
            raise ValueError(msg.format(op.name, name))
    return input_vars


def register_axis_update_op(cls=None, similar_ops=[]):
    """
    :param similar_ops: these ops share the same "update" and
    "can_transpose_pass" methods as the base class.
    For example: the class "transform_reduce_mean" corresponding to
    op "reduce_mean" can be shared with other ops such as
    "reduce_prod", "reduce_sum" etc
    """

    def class_wrapper(op_update_cls):
        cls_name = op_update_cls.__name__
        # remove "transform_" to get type of op
        op_type = cls_name[len("transform_") :]
        if op_type in AXIS_UPDATE_OPS:
            raise ValueError(
                "Update class for op '{}' already defined".format(op_update_cls)
            )
        AXIS_UPDATE_OPS[op_type] = op_update_cls
        for similar_op_type in similar_ops:
            if similar_op_type in AXIS_UPDATE_OPS:
                raise ValueError(
                    "Update class for op of type '{}' already defined".format(op_type)
                )
            AXIS_UPDATE_OPS[similar_op_type] = op_update_cls
        return op_update_cls

    if cls is None:
        return class_wrapper

    return class_wrapper


class transform_axis_update_ops(object):
    """
    Parent class for every axis update op's class

    An axis update op is an op that can be updated, such that it can allow a transpose layer to "pass" through it.
    That is,

    op(transpose(x)) == transpose(op_updated(x))

    where
    "op" : original op,
    "op_updated": op after being updated.

    Example:

    if x is a tensor of rank 2, and transpose has perm=[1,0],
    then

    reduce_mean[axis=1](transpose(x)) == transpose(reduce_mean[axis=0](x))

    here reduce_mean op with axis=1 can be updated to a reduce_mean op with axis=0,
    to allow the transpose to "pass" through it, i.e. get applied after it.

    """

    def __init__(self, op, transpose_axes):
        self.op = op
        self.transpose_axes = transpose_axes

    def can_transpose_pass(self):
        """
        Each "axis" op must determine whether it can act like a unary op
        and allow the transpose to pass through.
        Return True if it can allow the transpose to pass through, otherwise return False.

        :return: bool
        """
        raise NotImplementedError("This function must be implemented by each op")

    def update(self):
        """
        A method that updates some attribute of the axis op,
        based on the transpose axes value.
        This method only gets called if "can_transpose_pass" returns True.

        Update the op such that the output %i2 should be equal to %o2

        Before:
        %i_1 = transpose_op(%i_0, perm=transpose_axes)
        %i2 = op(%i1)

        After:
        %o1 = op_updated(%i0)
        %o2 = transpose_op(%o1, perm=transpose_axes)

        :return: None
        """
        raise NotImplementedError("This function must be implemented by each op")


@register_axis_update_op()
class transform_concat(transform_axis_update_ops):
    def __init__(self, **kwargs):
        super(transform_concat, self).__init__(**kwargs)
        self.axis_var = self.op.inputs["axis"]

    def can_transpose_pass(self):
        if self.axis_var.val is not None:
            return True
        return False

    def update(self):
        new_axis_val = self.transpose_axes[self.axis_var.val]
        inputs = list(self.op.inputs["values"])

        # to be used, if there is a consant inputs to the concat op
        transpose_perm_for_const = [0] * len(self.transpose_axes)
        for i, axis in enumerate(self.transpose_axes):
            transpose_perm_for_const[axis] = i

        # if there is a constant input, transpose it
        for input_var in inputs:
            if input_var.op.op_type == "const":
                const_val = input_var.val
                new_const_val = np.transpose(const_val, transpose_perm_for_const)
                # insert a new constant JUST before the op
                with self.op.enclosing_block:
                    new_const_input_var = mb.const(
                        val=new_const_val, mode="immediate_value", before_op=self.op
                    )
                self.op.enclosing_block.replace_uses_of_var_after_op(
                    anchor_op=new_const_input_var.op,
                    end_op=self.op,
                    old_var=input_var,
                    new_var=new_const_input_var,
                    no_check_var_types=True,
                )

        # insert a new constant for the new axis, JUST before the op
        with self.op.enclosing_block:
            new_axis_var = mb.const(
                val=new_axis_val, mode="immediate_value", before_op=self.op
            )

        self.op.enclosing_block.replace_uses_of_var_after_op(
            anchor_op=new_axis_var.op,
            end_op=self.op,
            old_var=self.axis_var,
            new_var=new_axis_var,
            no_check_var_types=True,
        )


@register_axis_update_op()
class transform_pad(transform_axis_update_ops):
    def __init__(self, **kwargs):
        super(transform_pad, self).__init__(**kwargs)
        self.pad_var = self.op.inputs["pad"]
        self.pad_op = self.pad_var.op
        self.mode = self.op.mode.val
        self.pad_amounts_new = None

    def _compute_new_pad_values(self):
        pad_amounts = np.reshape(self.pad_var.val, [-1, 2])
        rank_diff = len(self.transpose_axes) - pad_amounts.shape[0]
        self.pad_amounts_new = copy.deepcopy(pad_amounts)
        # append "rank_diff" rows of zeros to the top
        self.pad_amounts_new = np.concatenate(
            (np.zeros((2 * rank_diff)).reshape(-1, 2), self.pad_amounts_new)
        )
        self.pad_amounts_new = self.pad_amounts_new.astype(pad_amounts.dtype)
        pad_amounts = np.concatenate(
            (np.zeros((2 * rank_diff)).reshape(-1, 2), pad_amounts)
        )
        for i, axis in enumerate(self.transpose_axes):
            self.pad_amounts_new[axis][0] = pad_amounts[i][0]
            self.pad_amounts_new[axis][1] = pad_amounts[i][1]
        # get the top "rank_diff" rows
        top_rows = self.pad_amounts_new[:rank_diff, :]
        if not np.all(top_rows == 0):
            return False
        # cut "rank_diff" from the top
        self.pad_amounts_new = self.pad_amounts_new[rank_diff:, :]
        self.pad_amounts_new = self.pad_amounts_new.flatten()
        return True

    def can_transpose_pass(self):
        if (
            len(_get_input_vars(self.op, only_nonconst_vars=True)) != 1
            or self.pad_op.op_type != "const"
        ):
            return False
        if len(self.transpose_axes) < 2:
            return False
        if not self._compute_new_pad_values():
            return False
        # check that if mode is not constant, the updated padding
        # would stay limited to last 2 axes
        if self.mode != "constant" and not np.all(self.pad_amounts_new[:-4] == 0):
            return False
        return True

    def update(self):
        self._compute_new_pad_values()
        # insert a new constant for pad val, JUST before the op
        with self.op.enclosing_block:
            new_pad_var = mb.const(
                val=self.pad_amounts_new, mode="immediate_value", before_op=self.op
            )
        self.op.enclosing_block.replace_uses_of_var_after_op(
            anchor_op=new_pad_var.op,
            end_op=self.op,
            old_var=self.pad_var,
            new_var=new_pad_var,
            no_check_var_types=True,
        )


@register_axis_update_op(
    similar_ops=[
        "reduce_l1_norm",
        "reduce_l2_norm",
        "reduce_max",
        "reduce_log_sum",
        "reduce_log_sum_exp",
        "reduce_min",
        "reduce_prod",
        "reduce_sum",
        "reduce_sum_square",
    ]
)
class transform_reduce_mean(transform_axis_update_ops):
    def __init__(self, **kwargs):
        super(transform_reduce_mean, self).__init__(**kwargs)
        self.axes_var = self.op.inputs["axes"]
        self.axes_op = self.axes_var.op

    def can_transpose_pass(self):
        # allow transpose to push through it only if keep_dims are True since that doesn't change the rank
        if self.op.inputs["keep_dims"].val:
            if self.axes_op.op_type == "const":
                return True
        return False

    def update(self):
        # update axis of the op
        old_axes_val = self.axes_var.val
        new_axes_val = [0] * len(old_axes_val)
        for i, axis in enumerate(old_axes_val):
            new_axes_val[i] = self.transpose_axes[axis]

        # insert a new constant for the axis, JUST before the op
        with self.op.enclosing_block:
            new_axis_var = mb.const(
                val=new_axes_val, mode="immediate_value", before_op=self.op
            )

        self.op.enclosing_block.replace_uses_of_var_after_op(
            anchor_op=new_axis_var.op,
            end_op=self.op,
            old_var=self.axes_var,
            new_var=new_axis_var,
            no_check_var_types=True,
        )


@register_axis_update_op(similar_ops=["mul", "sub", "real_div", "maximum", "minimum"])
class transform_add(transform_axis_update_ops):
    def __init__(self, **kwargs):
        super(transform_add, self).__init__(**kwargs)

    def can_transpose_pass(self):
        const_input = None
        if self.op.inputs["x"].op and self.op.inputs["x"].op.op_type == "const":
            const_input = self.op.inputs["x"]
            other_input = self.op.inputs["y"]
        if self.op.inputs["y"].op and self.op.inputs["y"].op.op_type == "const":
            if const_input is not None:
                return False  # both inputs are constant
            const_input = self.op.inputs["y"]
            other_input = self.op.inputs["x"]
        if const_input is None:
            return True
        if not isinstance(const_input.val, (np.ndarray, np.generic)):
            return False
        rank_const_input = len(const_input.val.shape)
        rank_other_input = len(other_input.shape) if other_input.shape else 0
        if rank_const_input <= 1 and rank_other_input > 0:
            return True
        return False

    def update(self):
        if len(_get_input_vars(self.op, only_nonconst_vars=True)) == 2:
            # nothing to update
            return

        for input_var in _get_input_vars(self.op):
            if input_var.op and input_var.op.op_type == "const":
                const_input_var = input_var
                break

        const_value = const_input_var.val
        if len(const_value.shape) == 0:
            # const is a scalar, no need to modify it
            return

        rank = len(self.transpose_axes)
        new_shape = [1] * rank
        new_shape[self.transpose_axes[-1]] = const_value.shape[0]
        new_const_val = np.reshape(const_value, new_shape)

        # insert a new constant JUST before the op
        with self.op.enclosing_block:
            new_const_var = mb.const(
                val=new_const_val, mode=const_input_var.op.mode, before_op=self.op
            )

        self.op.enclosing_block.replace_uses_of_var_after_op(
            anchor_op=new_const_var.op,
            end_op=self.op,
            old_var=const_input_var,
            new_var=new_const_var,
            no_check_var_types=True,
        )


class HypotheticalValue(object):
    # A hypothetical value.
    # Simply wraps a Var.
    # Actual Var it wraps doesn't really matter, its mainly for debugging.
    # This class really exists to differentiate a "LazyTransposeHypotheticalValue" type with a
    # non-"LazyTransposeHypotheticalValue" type
    def __init__(self, var=None):
        self.value = var  # type : Var


class LazyTransposeHypotheticalValue(object):
    # a hypothetical value that represents a transpose op on top of a hypothetical value,
    # or a collection of transpose_ops, which have the same "perm" parameter

    def __init__(self, hypothetical_value, transpose_ops, perm):

        # Input hypothetical value to the transpose op.
        # When there are multiple transpose ops, this is the incoming hypothetical value to any one of those
        self.wrapped_hypothetical_value = hypothetical_value  # type : HypotheticalValue

        if not isinstance(hypothetical_value, HypotheticalValue):
            raise ValueError(
                "transpose optimization pass: incorrect type passed for hypothetical_value"
            )

        for op in transpose_ops:
            if op.op_type != "transpose":
                raise ValueError(
                    "transpose optimization pass: LazyTransposeHypotheticalValue can only be made with transpose ops"
                )
            perm_op = list(op.inputs["perm"].val)
            if perm_op != perm:
                raise ValueError(
                    "transpose optimization pass: LazyTransposeHypotheticalValue can only be made with transpose ops with the same 'perm' values"
                )

        self.perm = perm  # type : list[int], perm parameter of all the transpose ops
        self.transpose_ops = transpose_ops  # type : Set(op)


class TransposeOptimization(object):
    def __init__(self, block):
        self.block = block

        # for each var in the block, this dictionary stores the hypothetical value that is assigned to it during
        # graph traversal
        self.var_to_hypothetical_value = (
            {}
        )  # type : var : HypotheticalValue or LazyTransposeHypotheticalValue
        # start out by filling this dictionary with all the inputs of the block
        for _, input_var in block.inputs.items():
            self.var_to_hypothetical_value[input_var] = HypotheticalValue(input_var)

        # Dictionaries below are used to store transpose cancellation/fusion information.
        # These are filled during the traversal of the graph,
        # after which they are used by the `_apply_transform` method

        # transpose op to the list of transpose ops that are its compliments and can be cancelled away with it
        self.transpose_op_to_cancel_ops = defaultdict(
            lambda: []
        )  # type : op : List[op]

        # transpose op to the list of ops before which it has to materialize, i.e. the root transpose op
        #  can be moved downstream in the graph, as far as these materialize ops
        self.transpose_op_to_materialize_ops = defaultdict(
            lambda: []
        )  # type : op : List[Tuple(op, Var)]

        # list of the ops that need to be updated (either their axis parameter or one of their constant inputs)
        # if the transpose op is fused away or moved downstream in the graph
        self.transpose_op_to_axis_update_ops = defaultdict(
            lambda: []
        )  # type : op : List[op]

        # for book keeping
        self.ops_updated = set()
        self.materialized_ops_handled = set()
        self.transpose_ops_removed = set()

        # save the output sinks' information
        self.old_output_vars = []
        self.output_sink_ops = []

        # We modify the graph temporarily for outputs
        self._add_output_sinks()

    def _add_output_sinks(self):
        # We add a identity sink for all outputs.
        self.old_output_vars = {var: var.name for var in self.block.outputs}
        new_outputs = []
        output_sinks_var = {}
        for out_var in self.block.outputs:
            with self.block:
                if out_var not in output_sinks_var:
                    out_sink = mb.identity(x=out_var)
                    output_sinks_var[out_var] = out_sink
                else:
                    out_sink = output_sinks_var[out_var]
                new_outputs.append(out_sink)
                self.output_sink_ops.append(out_sink.op)
        self.block.set_outputs(new_outputs)

    def _visit_unary_like_op(self, op, input_var=None):
        # pass the input var's hypothetical_value to the output var's, since shape invariant ops do
        # not modify the incoming hypothetical_value

        if input_var is None:
            input_var = op.inputs["x"]

        if len(op.outputs) > 1:
            msg = (
                "transpose optimization pass: op '{}', of type = '{}', has multiple outputs, hence it"
                "cannot be handled like a unary op"
            )
            raise ValueError(msg.format(op.name, op.op_type))
        self.var_to_hypothetical_value[op.outputs[0]] = self.var_to_hypothetical_value[
            input_var
        ]

    def _visit_materialize_op(self, op):
        # this is the catch all category of ops
        # these are the "not-lazy-transpose-pass-through" kind of ops
        # output hypothetical_value is same as the vars
        for out_var in op.outputs:
            self.var_to_hypothetical_value[out_var] = HypotheticalValue(out_var)

        # check for the inputs
        # if there is a lazy transpose hypothetical value as an input,
        # all the transpose ops it hold,
        # need to be materialized here now, i.e., we should update "transpose_op_to_materialize_ops"
        for input_var in _get_input_vars(op):
            input_hypothetical_value = self.var_to_hypothetical_value[input_var]
            if isinstance(input_hypothetical_value, LazyTransposeHypotheticalValue):
                all_lazy_transpose_ops = input_hypothetical_value.transpose_ops
                for transpose_op in all_lazy_transpose_ops:
                    self.transpose_op_to_materialize_ops[transpose_op].append(
                        (op, input_var)
                    )

    def _visit_axis_update_op(self, op):
        """
        Check that all non constant inputs are of type LazyTransposeHypotheticalValue with the same perm value
        This check is common for all "axis update" ops.
        """
        input_vars = _get_input_vars(op, only_nonconst_vars=True)
        perm = None
        for i, var in enumerate(input_vars):
            hypothetical_value = self.var_to_hypothetical_value[var]
            if not isinstance(hypothetical_value, LazyTransposeHypotheticalValue):
                self._visit_materialize_op(op)
                return
            if i == 0:
                perm = hypothetical_value.perm
            elif perm != hypothetical_value.perm:
                self._visit_materialize_op(op)
                return

        # checks specific to the op type
        op_cls = AXIS_UPDATE_OPS.get(op.op_type, None)
        if op_cls is None:
            raise ValueError(
                "Transform class for op of type '{}' not found".format(op.op_type)
            )

        if not op_cls(**{"op": op, "transpose_axes": perm}).can_transpose_pass():
            self._visit_materialize_op(op)
            return

        # add this op to the dictionary "transpose_op_to_axis_update_ops"
        # and update self.var_to_hypothetical_value[op.outputs[0]]
        all_lazy_transpose_ops = set()
        for var in input_vars:
            input_hypothetical_value = self.var_to_hypothetical_value[var]
            all_lazy_transpose_ops.update(input_hypothetical_value.transpose_ops)

        for transpose_op in all_lazy_transpose_ops:
            self.transpose_op_to_axis_update_ops[transpose_op].append(op)

        self.var_to_hypothetical_value[op.outputs[0]] = LazyTransposeHypotheticalValue(
            input_hypothetical_value.wrapped_hypothetical_value,
            all_lazy_transpose_ops,
            perm,
        )

    def _visit_transpose_op(self, op):
        input_var = op.inputs["x"]
        if op.inputs["perm"].val is None:
            self._visit_materialize_op(op)
            return
        perm = list(op.inputs["perm"].val)
        input_hypothetical_value = self.var_to_hypothetical_value[input_var]

        """
        There are 3 cases to handle:

        1. input type == HypotheticalValue
        2. input type == LazyTransposeHypotheticalValue and this op is the transpose compliment of it
        3. input type == LazyTransposeHypotheticalValue and this op is NOT the transpose compliment of it
        """

        if isinstance(input_hypothetical_value, HypotheticalValue):
            # case 1
            # the input is not a lazy transpose.
            # Since the current node is a transpose, there are two sub-cases.
            #  a) It's a output node. We materialize it directly.
            #  b) It might get cancelled downstream, so make the output var's
            #     hypothetical_value a lazy transpose
            if op.outputs[0] in self.old_output_vars:
                self._visit_materialize_op(op)
            else:
                self.var_to_hypothetical_value[
                    op.outputs[0]
                ] = LazyTransposeHypotheticalValue(
                    input_hypothetical_value, set([op]), perm
                )
            return

        # input is a Lazy transpose hypothetical value. Lets first check whether the current
        # transpose cancels it or not
        do_cancel = _do_transposes_cancel(input_hypothetical_value.perm, perm)
        if do_cancel:
            # case 2
            # transposes cancel, so now the hypothetical_value of the output will
            # be same as the hypothetical value wrapped inside the upstream lazy transpose
            self.var_to_hypothetical_value[
                op.outputs[0]
            ] = input_hypothetical_value.wrapped_hypothetical_value
            # also update the dictionary "transpose_op_to_cancel_ops"
            all_lazy_transpose_ops = input_hypothetical_value.transpose_ops
            for transpose_op in all_lazy_transpose_ops:
                self.transpose_op_to_cancel_ops[transpose_op].append(op)
        else:
            # case 3
            # transposes don't cancel
            # this is same as a materialize op then
            self._visit_materialize_op(op)

    def _visit_op(self, op):

        input_vars = _get_input_vars(op)
        for var in input_vars:
            assert (
                var in self.var_to_hypothetical_value
            ), "transpose optimization pass: hypothetical value for var '{}', not found".format(
                var.name
            )

        if op in self.output_sink_ops:
            self._visit_materialize_op(op)
        elif op.op_type in UNARY_LIKE_OP_TYPES:
            self._visit_unary_like_op(op)
        elif op.op_type in AXIS_UPDATE_OPS:
            self._visit_axis_update_op(op)
        elif op.op_type == "transpose":
            self._visit_transpose_op(op)
        elif op.op_type == "const":
            self.var_to_hypothetical_value[op.outputs[0]] = HypotheticalValue(
                op.outputs[0]
            )
        else:
            self._visit_materialize_op(op)

    def block_traversal(self):

        # Since the ops are already organized in a topological manner,
        # simply iterate through all the ops

        for op in self.block.operations:
            self._visit_op(op)


    def _verify_cancellable_transposes(self):

        # invert "transpose_op_to_cancel_ops"
        transpose_cancel_ops_to_starting_transpose_set = defaultdict(lambda: set())
        for op, cancel_ops_list in self.transpose_op_to_cancel_ops.items():
            for cancel_op in cancel_ops_list:
                transpose_cancel_ops_to_starting_transpose_set[cancel_op].update(
                    set([op])
                )

        for op in transpose_cancel_ops_to_starting_transpose_set:
            assert (
                op not in self.transpose_op_to_cancel_ops
            ), "transpose reduction optimization: transpose op '{}' cannot be both a starting and cancel op".format(
                op.name
            )

        # invert "transpose_op_to_materialize_ops"
        materizalize_ops_to_starting_transpose_set = defaultdict(lambda: set())
        for op, materialize_ops in self.transpose_op_to_materialize_ops.items():
            for materialize_op, edge in materialize_ops:
                materizalize_ops_to_starting_transpose_set[materialize_op].update(
                    set([op])
                )

                # the starting transpose op may not be in "transpose_op_to_cancel_ops"
                # but it needs to be removed if it materializes later, hence we need to add it
                # to the "transpose_op_to_cancel_ops", with an empty value, i.e. no other ops to cancel because of it
                if op not in self.transpose_op_to_cancel_ops:
                    self.transpose_op_to_cancel_ops[op] = []

        # (starting transpose ops) and (transpose cancel ops + materialize ops) form a bipartite graph.
        # Find the connected components of this graph, by doing a BFS traversal
        connected_components = []  # List[(Set(op), Set(op)), Set(op)]
        visited = {}
        for op in list(self.transpose_op_to_cancel_ops.keys()):
            if op in visited:
                continue
            visited[op] = 1
            set_a = set([op])  # set of starting transpose ops
            set_b1 = set()  # set of transpose cancel ops connected to set_a
            set_b2 = set()  # set of materialize ops connected to set_a
            queue = []
            queue.extend(self.transpose_op_to_cancel_ops[op])
            if op in self.transpose_op_to_materialize_ops:
                materialize_ops_list = list(
                    list(zip(*self.transpose_op_to_materialize_ops[op]))[0]
                )
                queue.extend(materialize_ops_list)
            while len(queue) > 0:
                o = queue.pop(0)
                visited[o] = 1
                # enque nodes connected to o
                if o in self.transpose_op_to_cancel_ops:
                    set_a.update(set([o]))
                    for neighbor_op in self.transpose_op_to_cancel_ops[o]:
                        if neighbor_op not in visited:
                            queue.append(neighbor_op)
                    if o in self.transpose_op_to_materialize_ops:
                        materialize_ops_list = list(
                            list(zip(*self.transpose_op_to_materialize_ops[o]))[0]
                        )
                        for neighbor_op in materialize_ops_list:
                            if neighbor_op not in visited:
                                queue.append(neighbor_op)
                elif o in transpose_cancel_ops_to_starting_transpose_set:
                    set_b1.update(set([o]))
                    for neighbor_op in transpose_cancel_ops_to_starting_transpose_set[
                        o
                    ]:
                        if neighbor_op not in visited:
                            queue.append(neighbor_op)
                else:
                    set_b2.update(set([o]))
                    for neighbor_op in materizalize_ops_to_starting_transpose_set[o]:
                        if neighbor_op not in visited:
                            queue.append(neighbor_op)
            connected_components.append((set_a, set_b1, set_b2))

        starting_ops_to_remove = (
            set()
        )  # starting ops to remove from the optimization list

        # now for each connected component, make a decision whether to cancel it or not
        # (either all transpose ops in a set get cancelled or they don't)
        for op_set, op_cancel_set, materialize_op_set in connected_components:

            block_output = False
            # check that output is not directly connected to a starting transpose op
            for op in op_set:
                if op.outputs[0] in self.block.outputs:
                    starting_ops_to_remove.update(op_set)
                    block_output = True
                    break
            if block_output:
                continue

            materizalize_set = set(list(materialize_op_set))
            if len(materizalize_set) >= len(op_set) + len(op_cancel_set):
                starting_ops_to_remove.update(op_set)

        # remove ops
        for op in starting_ops_to_remove:
            self.transpose_op_to_cancel_ops.pop(op, None)

    def _remove_transpose_ops(self, starting_transpose_op):

        perm = list(starting_transpose_op.inputs["perm"].val)
        starting_transpose_op_out_var = starting_transpose_op.outputs[0]
        starting_transpose_op_input_var = starting_transpose_op.inputs["x"]

        # update all the "axis_update" ops
        for op in self.transpose_op_to_axis_update_ops.get(starting_transpose_op, []):
            if op not in self.ops_updated:
                op_cls = AXIS_UPDATE_OPS.get(op.op_type, None)
                op_cls(**{"op": op, "transpose_axes": perm}).update()
                self.ops_updated.add(op)

        # short circuit starting_transpose_op and its cancel ops

        to_be_removed_ops = []
        name_changed_vars = set()

        for op in [starting_transpose_op] + self.transpose_op_to_cancel_ops[
            starting_transpose_op
        ]:
            if op in self.transpose_ops_removed:
                continue

            to_be_removed_ops.append(op)
            self.transpose_ops_removed.add(op)

            input_var = op.inputs["x"]  # input to the transpose op
            output_var = op.outputs[0]  # output of the transpose op
            parent_op = input_var.op  # parent op of the transpose op

            if output_var in self.old_output_vars:
                # output is a block output, so this must be one of the "edge" transpose compliment ops
                # We need to set `input_var` as the block output var
                # Change the name of the input_var to match the block output if input_var is not changed.
                # If the same input_var is in output twice, we can't rename it twice, therefore we initiate an
                # Identity op to match the name
                if input_var in self.block.inputs.values():
                    with self.block:
                        input_var = mb.identity(x=input_var, before_op=op, name=output_var.name)
                        parent_op = None  # set anchor op as None.
                elif input_var not in name_changed_vars:
                    input_var.name = output_var.name
                    input_var.op.name = output_var.op.name
                    name_changed_vars.update([input_var])
                else:
                    with self.block:
                        input_var = mb.identity(x=input_var, before_op=op, name=output_var.name)
                        parent_op = input_var.op

            # connect all the child ops of the output_var to the parent of the transpose op.
            self.block.replace_uses_of_var_after_op(
                anchor_op=parent_op,
                old_var=output_var,
                new_var=input_var,
                no_check_var_types=True,
            )

        """
        Insert a transpose op JUST before each one of the materialize ops
        i.e.
        Given:  %i1 = op(...)
                ...
                ... = materialize_op(..., %i1 ,...)
                ...

        Result: %i1 = op(...)
                ...
                %i2 = transpose_op(%i1, %perm)
                ... = materialize_op(..., %i2 ,...)
                ...
        """
        for op, input_var in self.transpose_op_to_materialize_ops.get(
            starting_transpose_op, []
        ):
            if (op, input_var) in self.materialized_ops_handled:
                continue

            self.materialized_ops_handled.add((op, input_var))
            with self.block:
                if input_var == starting_transpose_op_out_var:
                    # materialize op is connected to the starting transpose op
                    # in this case, connect to its parent
                    if op in self.output_sink_ops:
                        continue
                    i1 = starting_transpose_op_input_var
                else:
                    i1 = input_var

                if op in self.output_sink_ops:
                    # The input_var of output sink is itself a output. We can safely
                    # modify the name of the input_var since it should only be consumed
                    # by block output here.
                    if i1 not in name_changed_vars:
                        x = mb.transpose(x=i1, perm=perm, before_op=op, name=i1.name)
                        i1.name = '_before_transpose_op_' + x.op.name
                        i1.op.name = '_before_transpose_op_' + x.op.name
                    else:
                        x = mb.transpose(x=i1, perm=perm, before_op=op, name=self.old_output_vars[i1])
                else:
                    x = mb.transpose(x=i1, perm=perm, before_op=op)

            self.block.replace_uses_of_var_after_op(
                anchor_op=x.op,
                end_op=op,
                old_var=i1,
                new_var=x,
                no_check_var_types=True,
            )

        self.block.remove_ops(to_be_removed_ops)

    def apply_transform(self):
        """
        Take in the data collected during graph traversal
        and transform the graph by cancelling out transpose ops that can be removed.
        """

        logging.debug(
            "Block before optimize transpose transform:\n{}".format(self.block)
        )
        if DEBUG:
            import graphviz

            graphviz.Source(
                self.block.get_dot_string(
                    highlight_debug_op_names=[], highlight_debug_op_types=["transpose"]
                )
            ).view(filename="/tmp/block_before_reduce_transpose")

        """
        First check which transposes can be cancelled.
        After this function call we get an updated dictionary "transpose_op_to_cancel_ops"
        with only the transpose ops that can really be cancelled in the graph
        Reasons to not cancel:
        - materialize_ops are greater than cancel_ops, so removing transpose will instead end up increasing the count of transposes
        - removing a transpose op can only be successful, if all of its cancel ops are removed, removing all the cancel ops
          is only successful if all of their starting transpose ops are removed and so on. This check is also done in
           "_verify_cancellable_transposes()"
        """
        self._verify_cancellable_transposes()

        # apply transform
        for transpose_op in self.transpose_op_to_cancel_ops:
            self._remove_transpose_ops(transpose_op)
        self.block.set_outputs([sink_op.x for sink_op in self.output_sink_ops])
        self.block.remove_ops(list(self.output_sink_ops))

        if DEBUG:
            graphviz.Source(
                self.block.get_dot_string(
                    highlight_debug_op_names=[], highlight_debug_op_types=["transpose"]
                )
            ).view(filename="/tmp/block_after_reduce_transpose")

        logging.debug(
            "Block after optimize transpose transform:\n{}".format(self.block)
        )

        for op in self.block.operations:
            op.type_value_inference(overwrite_output=True)


def reduce_transposes_block(block):
    """
    Only apply the optimization if the block is flat,
    i.e, it does not contain any op which contains a sub-block.
    TODO:
    Removing transposes and transpose compliments requires re-running
    type inference for the set of ops in between the fused transpose ops,
    which is simpler to do when all the ops in the block are free of sub blocks.
    The case of transpose fusion with sub-block containing ops needs to be handled with more care and test cases.
    """
    for op in list(block.operations):
        if len(op.blocks) > 0:
            return

    opt_transposes = TransposeOptimization(block)
    opt_transposes.block_traversal()
    opt_transposes.apply_transform()


@register_pass(namespace="common")
def reduce_transposes(prog):
    for f_name, f in prog.functions.items():
        reduce_transposes_block(f)
